#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# pylint: disable=invalid-name,line-too-long,too-many-locals,consider-using-sys-exit

import operator
import re
import time
import argparse
import sys
import os
import itertools
import pandas as pd
import pickle
import sklearn
import sklearn.svm
import sklearn.tree
import sklearn.preprocessing
import numpy as np
import sklearn.metrics
import sklearn.impute
import matplotlib.pyplot as plt
import sklearn.ensemble
import sklearn.linear_model
import helper
import xgboost as xgb
import lightgbm as lgbm
import ast
import math
import functools
try:
    import mlflow
except ImportError:
    mlflow_avail = False
else:
    mlflow_avail = True

ver = sklearn.__version__.split(".")
print("version: ", ver)
if int(ver[0]) == 0 and int(ver[1]) < 20:
    from sklearn.cross_validation import train_test_split
else:
    from sklearn.model_selection import train_test_split

MISSING=np.NaN

class MyEnsemble:
    def __init__(self, models):
        self.models = models

    def fit(self, X_train, y_train, weights=None):
        assert weights is None
        for model in self.models:
            model.fit(X_train, y_train)

    def predict(self, X_data):
        #df = pd.DataFrame(index=range(numRows),columns=range(numCols))
        vals = np.ndarray(shape=(len(self.models), len(X_data)), dtype=float)
        for i in range(len(self.models)):
            pred = self.models[i].predict(X_data)
            vals[i] = pred

        ret = np.median(vals, axis=0)
        return ret



class Learner:
    def __init__(self, df):
        self.df = df

    def count_bad_ok(self, df):
        files = df[["x.class", "rdb0.dump_no"]].groupby("x.class").count()
        if files["rdb0.dump_no"].index[0] == 0:
            bad = files["rdb0.dump_no"][0]
            ok = files["rdb0.dump_no"][1]
        else:
            bad = files["rdb0.dump_no"][1]
            ok = files["rdb0.dump_no"][0]

        assert bad > 0, "No need to train, data only contains BAD(0)"
        assert ok > 0, "No need to train, data only contains OK(1)"

        return bad, ok

    def filter_percentile(self, df, features, perc):
        low = df.quantile(perc, axis=0)
        high = df.quantile(1.0-perc, axis=0)
        df2 = df.copy()
        for i in features:
            df2 = df2[(df2[i] >= low[i]) & (df2[i] <= high[i])]
            print("Filtered to %f on %-30s, shape now: %s" %
                  (perc, i, df2.shape))

        print("Original size:", df.shape)
        print("New size:", df2.shape)
        return df2

    def filtered_conf_matrixes(self, dump_no, data, features, to_predict, clf,
                               toprint, highlight=False):
        # filter test data
        if dump_no is not None:
            print("\nCalculating confusion matrix -- dump_no == %s" % dump_no)
            toprint += " dump no %d" % dump_no
            data2 = data[data["rdb0.dump_no"] == dump_no]
        else:
            print("\nCalculating confusion matrix -- ALL dump_no")
            data2 = data

        return helper.calc_regression_error(
            data2, features, to_predict, clf, toprint,
            highlight=highlight)

    def importance_XGB(self, clf, features):
        impdf = []
        #print("clf:", clf)
        #print("clf-booster:", clf.feature_importances_)

        for i in range(len(clf.feature_importances_)):
            score = clf.feature_importances_[i]
            ft = features[i]
            impdf.append({'feature': ft, 'importance': score})
        impdf = pd.DataFrame(impdf)
        impdf = impdf.sort_values(by='importance', ascending=False).reset_index(drop=True)
        impdf['importance'] /= impdf['importance'].sum()
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)
        pd.set_option('display.width', None)
        pd.set_option('display.max_colwidth', None)
        print("impdf:", impdf)

        return impdf

    def dump_ml_test_data(self, test, fname, to_predict) :
        f = open(fname, "w")
        test.reset_index(inplace=True)
        for i in range(test.shape[0]):
            towrite = ""
            towrite += "%s " % test[TODO_go_through_all_features_here].iloc[i]
            assert False, "not implemented, must put all features here in a loop"
            towrite += "%s " % test[to_predict].iloc[i]
            towrite += "\n"
            f.write(towrite)

    def scale_and_impute(self, train, test, features, extra_feats):
        trans_train = train[features].values
        trans_test = test[features].values

        # Scale data if linear
        my_scaler = sklearn.preprocessing.Normalizer(
            norm='l1',
            copy=False)
        my_scaler.fit_transform(trans_train)
        my_scaler.transform(trans_test)

        # Impute data
        imp_mean = sklearn.impute.SimpleImputer(
            #missing_values=MISSING,
            copy=False,
            strategy='mean')
        imp_mean.fit_transform(trans_train)
        imp_mean.transform(trans_test)

        # recreate dataframe
        trans_train = np.append(trans_train, train[extra_feats].values, axis=1)
        df_trans_train = pd.DataFrame(trans_train, columns=features+extra_feats)
        train = df_trans_train

        trans_test = np.append(trans_test, test[extra_feats].values, axis=1)
        df_trans_test = pd.DataFrame(trans_test, columns=features+extra_feats)
        test = df_trans_test

        print("Value distribution of 'rdb0.glue' in test:\n%s" % test["rdb0.glue"].value_counts())
        print("Value distribution of 'rdb0.glue' in train:\n%s" % train["rdb0.glue"].value_counts())

        return train, test

    def one_regressor(self, features, to_predict):
        print("-> Number of features  :", len(features))
        print("-> Number of datapoints:", self.df.shape)
        print("-> Predicting          :", to_predict)

        # these are needed for prediction/later checks, so let's add them in
        #       if they are not already in the features
        extra_feats = [to_predict]
        for missing_needed in ["rdb0.glue", "rdb0.dump_no"]:
            if missing_needed not in features:
                extra_feats.append(missing_needed)
        df = self.df[features+extra_feats].copy()
        if options.verbose:
            pd.set_option('display.max_rows', len(df.dtypes))
            print(df.dtypes)
            pd.reset_option('display.max_rows')


        if options.check_row_data:
            helper.check_too_large_or_nan_values(df, features)

        print("Value distribution of 'rdb0.dump_no':\n%s" % df["rdb0.dump_no"].value_counts())
        print("Value distribution of 'rdb0.glue':\n%s" % df["rdb0.glue"].value_counts())
        print("Value distribution of to_predict:\n%s" % df[to_predict].value_counts())

        train, test = train_test_split(df, test_size=0.33, random_state=prng)
        if options.regressor in ["linear", "tree"]:
            train, test = self.scale_and_impute(train, test, features, extra_feats)


        if options.poly_features:
            transformed = df[features].values

            # poly transform
            poly = sklearn.preprocessing.PolynomialFeatures(2)
            transformed = poly.fit_transform(transformed)
            features = poly.get_feature_names(input_features=features)

            # recreate dataframe
            transformed = np.append(transformed, df[extra_feats].values, axis=1)
            df_trans = pd.DataFrame(transformed, columns=features+extra_feats)
            df = df_trans

        X_train = train[features]
        y_train = train[to_predict]

        t = time.time()
        clf = None

        split_point = helper.calc_min_split_point(
            df, options.min_samples_split)

        clf_tree = sklearn.tree.DecisionTreeRegressor(
            max_depth=options.xboost_max_depth,
            min_samples_split=split_point,
            random_state=prng)
        clf_svm_pre = sklearn.svm.SVC(
            C=500,
            gamma=10**-5,
            random_state=prng)
        clf_svm = sklearn.ensemble.BaggingClassifier(
            clf_svm_pre,
            n_estimators=3,
            max_samples=0.5, max_features=0.5,
            random_state=prng)
        clf_linear = sklearn.linear_model.LinearRegression()
        clf_forest = sklearn.ensemble.RandomForestRegressor(
            n_estimators=options.num_trees,
            max_depth=options.xboost_max_depth,
            max_features="sqrt",
            #min_samples_leaf=split_point,
            random_state=prng)
        clf_ridge = sklearn.linear_model.Ridge(alpha=.5)
        clf_lasso = sklearn.linear_model.Lasso()
        clf_elasticnet = sklearn.linear_model.ElasticNet()
        clf_xgboost = xgb.XGBRegressor(
                objective='reg:squarederror',
                #missing=MISSING,
                min_child_weight=options.min_child_weight_xgboost, # from doc: "In linear regression task, this simply corresponds to minimum number of instances needed to be in each node."
                max_depth=options.xboost_max_depth,
                subsample=options.xgboost_subsample,
                n_estimators=options.n_estimators_xgboost)
        clf_lgbm = model = lgbm.LGBMRegressor(
            subsample=options.xgboost_subsample,
            min_child_samples=options.min_child_weight_xgboost, # from doc: "Minimum number of data needed in a child"
            max_depth=options.xboost_max_depth,
            n_estimators=options.n_estimators_xgboost)

        if options.regressor == "tree":
            clf = clf_tree
        elif options.regressor == "svm":
            clf = clf_svm
        elif options.regressor == "linear":
            clf = clf_linear
        elif options.regressor == "ridge":
            clf = clf_ridge
        elif options.regressor == "lasso":
            clf = clf_lasso
        elif options.regressor == "elasticnet":
            clf = clf_elasticnet
        elif options.regressor == "forest":
            clf = clf_forest
        elif options.regressor == "lgbm":
            clf = clf_lgbm
        elif options.regressor == "xgb":
            print("Using xgboost no. estimators:", options.n_estimators_xgboost)
            clf = clf_xgboost
        elif options.regressor == "median":
            mylist = [("xgb", clf_xgboost), ("linear", clf_linear),
                      ("elasticnet", clf_elasticnet), ("lasso", clf_lasso)]
            clf = sklearn.ensemble.VotingRegressor(estimators=mylist, weights=[1.0, 0.5, 0.5, 0.5])
            #clf = MyEnsemble(mylist)
        else:
            print(
                "ERROR: You MUST give one of: tree/forest/svm/linear/bagging classifier")
            exit(-1)
        if options.gen_topfeats:
            print("WARNING: Replacing normal forest/xgboost with top feature computation!!!")
            if options.regressor == "forest":
                print("for TOP calculation, we are replacing the forest predictor!!!")
                clf = sklearn.ensemble.RandomForestRegressor(
                    n_estimators=options.num_trees*5,
                    max_features="sqrt",
                    random_state=prng)
            elif options.regressor == "xgb":
                print("for TOP calculation, we are replacing the XGBOOST predictor!!!")
                clf = xgb.XGBRegressor(
                    objective='reg:squarederror',
                    min_child_weight=options.min_child_weight_xgboost,
                    max_depth=options.xboost_max_depth,
                    #missing=MISSING
                    )
            else:
                print("Error: --topfeats only works with xgboost/forest")
                exit(-1)


        clf.fit(X_train, y_train)
        print("Training finished. T: %-3.2f" % (time.time() - t))

        if mlflow_avail:
            mlflow.log_param("features used", features)
            # mlflow.log_metric("all features: ", train.columns.values.flatten().tolist())
            mlflow.log_metric("train num rows", train.shape[0])
            mlflow.log_metric("test num rows", test.shape[0])


        if options.dot is not None:
            if options.regressor == "tree":
                helper.output_to_classical_dot(
                    clf, features,
                    fname="{name}-{table}-{tier}.dot".format(
                        name=options.dot, tier=options.tier, table=options.table))

            elif options.regressor == "xgb":
                dot_data = xgb.to_graphviz(booster=clf, num_trees=9)
                with open("x.dot", "w") as f:
                    f.write("%s" % dot_data)

            else:
                print("ERROR: You cannot use the DOT function on non-trees")
                exit(-1)

        if options.basedir:
            fname_pred_out = options.basedir + "/predictor-{table}-{tier}-{regr}.json".format(
                tier=options.tier, table=options.table, regr=options.regressor)
            if options.regressor == "xgb":
                booster = clf.get_booster()
                booster.save_model(fname_pred_out)
                print("==> Saved XGB model to: ", fname_pred_out)
            elif options.basedir and options.regressor == "lgbm":
                clf.booster_.save_model(fname_pred_out)
                print("==> Saved LGBM model to: ", fname_pred_out)
        else:
            print("WARNING: NOT writing code -- you must use xgboost/lgbm and give dir for that")


        # print feature rankings
        if options.regressor == "forest":
            helper.print_feature_ranking(
                clf, X_train,
                top_num_features=200,
                features=features,
                plot=options.show)
        elif options.regressor == "xgb":
            self.importance_XGB(clf, features=features)

        # print distribution of error
        print("--------------------------")
        print("-       test data        -")
        print("--------------------------")
        for dump_no in [1, 2, 3, 10, 20, 40, None]:
            self.filtered_conf_matrixes(
                dump_no, test, features, to_predict, clf, "test data", highlight=True)
        print("--------------------------------")
        print("--      train+test data        -")
        print("--------------------------------")
        for dump_no in [1, None]:
            self.filtered_conf_matrixes(
                dump_no, pd.concat([test, train]), features, to_predict, clf, "test and train data")
        print("--------------------------")
        print("-       train data       -")
        print("--------------------------")
        self.filtered_conf_matrixes(
            dump_no, train, features, to_predict, clf, "train data")

        #print(test[features+to_predict])
        #print(test[to_predict])
        if options.dump_example_data:
            with open("example-data.dat", "wb") as f:
                pickle.dump(test[features+[to_predict]], f)
            self.dump_ml_test_data(test, "../ml_perf_test.txt-{table}-{tier}".format(
                table=options.table, tier=options.tier))
            print("Example data dumped")

    def rem_features(self, feat, to_remove):
        print("To remove: " , to_remove)
        feat_less = list(feat)
        for rem in to_remove:
            for feat in list(feat_less):
                if rem in feat:
                    if options.verbose:
                        print("Removing due to ", rem, " feature from feat_less:", feat)
                    feat_less.remove(feat)

        print("Done.")
        return feat_less

    def learn(self):
        if options.raw_data_plots:
            pd.options.display.mpl_style = "default"
            df.hist()
            df.boxplot()

        if options.features != "best_only":
            features = list(self.df)

            # remove features that would be "cheating" or useless
            torem = []
            for table in ["used_later", "used_later_anc"]:
                for tier in ["short", "long", "forever"]:
                    torem. append("x.{table}_{tier}".format(tier=tier, table=table))

            torem.extend([
                "x.class",
                "x.a_lifetime",
                "fname",
                "sum_cl_use.num_used",
                "x.sum_cl_use",
                "rdb0.dump_no",
                "fname"])
            features = self.rem_features(features, torem)
        else:
            del df["fname"]
            features = helper.get_features(options.best_features_fname)

        to_predict = "x.{table}_{tier}".format(tier=options.tier, table=options.table)
        self.one_regressor(features, to_predict)


if __name__ == "__main__":
    usage = "usage: %(prog)s [options] file.pandas"
    parser = argparse.ArgumentParser(usage=usage)

    parser.add_argument("fname", type=str, metavar='PANDASFILE')
    parser.add_argument("--seed", default=None, type=int,
                        dest="seed", help="Seed of PRNG")
    parser.add_argument("--verbose", "-v", action="store_true", default=False,
                        dest="verbose", help="Print more output")
    parser.add_argument("--printfeat", action="store_true", default=False,
                        dest="print_features", help="Print features")
    parser.add_argument("--features", default="best_only", type=str, dest="features",
                        help="What features to use: all_computed, best_only, best_also, no_computed ")
    parser.add_argument("--bestfeatfile", type=str, default="../../scripts/crystal/best_features-rdb0-only.txt",
                        dest="best_features_fname", help="Name and position of best features file that lists the best features in order")
    parser.add_argument("--polyfeats", action="store_true", default=False,
                        dest="poly_features", help="Add polynomial features")
    parser.add_argument("--dat", type=str, default=None,
                        dest="dat_file", help="Output Pickle of dataframe here")
    parser.add_argument("--dumpexample", default=False, action="store_true",
                        dest="dump_example_data", help="Dump example data")

    # tree/forest options
    parser.add_argument("--split", default=0.01, type=float, metavar="RATIO",
                        dest="min_samples_split", help="Split in tree if this many samples or above. Used as a percentage of datapoints")
    parser.add_argument("--numtrees", default=100, type=int,
                        dest="num_trees", help="How many trees to generate for the forest")

    # generation of predictor
    parser.add_argument("--dot", type=str, default=None,
                        dest="dot", help="Create DOT file")
    parser.add_argument("--filterdot", default=0.05, type=float,
                        dest="filter_dot", help="Filter the DOT output from outliers so the graph looks nicer")
    parser.add_argument("--show", action="store_true", default=False,
                        dest="show", help="Show visual graphs")
    parser.add_argument("--check", action="store_true", default=False,
                        dest="check_row_data", help="Check row data for NaN or float overflow")
    parser.add_argument("--checkverbose", action="store_true", default=False,
                        dest="check_row_data_verbose", help="Check row data for NaN or float overflow in a verbose way, printing them all")
    parser.add_argument("--rawplots", action="store_true", default=False,
                        dest="raw_data_plots", help="Display raw data plots")
    parser.add_argument("--basedir", type=str,
                        dest="basedir", help="The base directory of where the CryptoMiniSat source code is")

    # data filtering
    parser.add_argument("--only", default=1.00, type=float,
                        dest="only_perc", help="Only use this percentage of data")

    # final generator top/final
    parser.add_argument("--topfeats", default=False, action="store_true",
                        dest="gen_topfeats", help="Only generate final predictor")

    # type of regressor
    parser.add_argument("--regressor", type=str, default="xgb",
                        dest="regressor", help="Final classifier should be a: tree, svm, linear, forest, xgboost, bagging")
    parser.add_argument("--xgboostestimators", default=10, type=int,
                        dest="n_estimators_xgboost", help="Number of estimators for xgboost")
    parser.add_argument("--xgboostminchild", default=10, type=int,
                        dest="min_child_weight_xgboost", help="Number of elements in the leaf to split it in xgboost")
    parser.add_argument("--xboostmaxdepth", default=10, type=int,
                        dest="xboost_max_depth", help="Max depth of xboost trees")
    parser.add_argument("--xgboostsubsample", default=1.0, type=float,
                        dest="xgboost_subsample", help="Subsample xgboost on each iteration")

    # which one to generate
    parser.add_argument("--tier", default=None, type=str,
                        dest="tier", help="Tier to do")
    parser.add_argument("--table", default="used_later", type=str,
                        dest="table", help="Table to do")

    options = parser.parse_args()
    prng = np.random.RandomState(options.seed)

    if options.fname is None:
        print("ERROR: You must give the pandas file!")
        exit(-1)

    assert options.min_samples_split <= 1.0, "You must give min_samples_split that's smaller than 1.0"
    if not os.path.isfile(options.fname):
        print("ERROR: '%s' is not a file" % options.fname)
        exit(-1)

    if options.tier is None:
        print("ERROR: you must set --tier, exiting")
        exit(-1)

    if options.table is None:
        print("ERROR: you must set --table, exiting")
        exit(-1)

    if options.best_features_fname is None and "best" in options.features:
        print("You must give best features filename or we cannot add best features")
        exit(-1)

    # ------------
    #  Log all parameters
    # ------------
    if mlflow_avail:
        mlflow.log_param("gen_topfeats", options.gen_topfeats)
        mlflow.log_param("tier", options.tier)
        mlflow.log_param("regressor", options.regressor)
        mlflow.log_param("basedir", options.basedir)
        mlflow.log_param("only_percentage", options.only_perc)
        mlflow.log_param("min_samples_split", options.min_samples_split)
        mlflow.log_param("xboost_max_depth", options.xboost_max_depth)
        mlflow.log_param("num_trees", options.num_trees)
        mlflow.log_param("features", options.features)
        mlflow.log_artifact(options.fname)

    # Read in Pandas Dataframe
    print("Reading dataframe....")
    df = pd.read_pickle(options.fname)

    print("Applying only...")
    df_tmp = df.sample(frac=options.only_perc, random_state=prng)
    df_before_dtype_conv = pd.DataFrame(df_tmp)
    del df_tmp
    del df
    print("-> Number of datapoints after applying '--only':", df_before_dtype_conv.shape)

    # We must convert these or we'll have trouble with inf, -inf, NaN for NULLs
    print("Converting datatypes to those supporting np.NA ...")
    if options.verbose:
        print("Datatypes before:")
        helper.print_datatypes(df_before_dtype_conv)
    df = df_before_dtype_conv.convert_dtypes(
        convert_integer=False, convert_string=False,
        convert_floating=False)
    del df_before_dtype_conv
    if options.verbose:
        print("Datatypes after:")
        helper.print_datatypes(df)

    # Check feature type sanity
    # only "fname" is allowed to be an object (a string)
    for name,ty in zip(list(df), df.dtypes):
        if name == "fname":
            assert ty == object
        else:
            if ty == object:
                print("name: " , name, " is object!")
            assert ty != object

    if options.print_features:
        for f in sorted(list(df)):
            print(f)

    # make missing (None) into NaN
    helper.make_missing_into_nan(df)

    # feature manipulation
    if options.features =="all_computed":
        helper.cldata_add_computed_features(df, options.verbose)
    elif options.features == "best_only" or options.features == "best_also":
        helper.add_features_from_fname(df, options.best_features_fname)
    elif options.features == "no_computed":
        helper.cldata_add_minimum_computed_features(df, options.verbose)
    else:
        print("ERROR: Unrecognized --features option!")
        exit(-1)

    # Check feature type sanity
    for name, mytype in df.dtypes.items():
        if str(mytype) == str("Int64") or str(mytype) == str("Float64"):
            assert False

    print("Filling NA with MISSING..")
    df.replace([np.inf, np.NaN, np.inf, np.NINF, np.Infinity], MISSING, inplace=True)

    if options.dat_file is not None:
        cols = list(df)
        df.to_pickle(options.dat_file, protocol=3)
        print("Dumped DF to pickle: ", options.dat_file)

    if options.check_row_data_verbose:
        helper.check_too_large_or_nan_values(df, list(df))

    # do the heavy lifting
    learner = Learner(df)
    learner.learn()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import print_function
import re


class XorToCNF:
    def __init__(self):
        self.cutsize = 4

    def get_max_var(self, clause):
        maxvar = 0

        tmp = clause.strip()
        if len(tmp) == 0:
            return 0

        assert re.search(r'^x? *-?\d+', tmp)

        if tmp[0] == 'x':
            tmp = tmp[1:]

        for lit in tmp.split():
            var = abs(int(lit))
            maxvar = max(var, maxvar)

        return maxvar

    def convert(self, infilename, outfilename, num_extra_cls=0):
        assert isinstance(self.cutsize, int)
        if self.cutsize <= 2:
            print("ERROR: The cut size MUST be larger or equal to 3")
            exit(-1)

        maxvar, numcls, extravars_needed, extracls_needed = self.get_stats(infilename)
        fout = open(outfilename, "w")
        fout.write("p cnf %d %d\n" %
                   (maxvar + extravars_needed,
                    numcls + extracls_needed + num_extra_cls))
        fin = open(infilename, "r")
        atvar = maxvar
        for line in fin:
            line = line.strip()

            # skip empty line
            if len(line) == 0:
                continue

            # skip header and comments
            if line[0] == 'c' or line[0] == 'p':
                continue

            if line[0] == 'x':
                # convert XOR to normal(s)
                xorclauses, atvar = self.cut_up_xor_to_n(line, atvar)
                for xorcl in xorclauses:
                    cls = self.xor_to_cnf_simple(xorcl)
                    for cl in cls:
                        fout.write(cl + "\n")
            else:
                # simply print normal clause
                fout.write(line + "\n")

        assert atvar == maxvar + extravars_needed
        fout.close()
        fin.close()

    def popcount(self, x):
        return bin(x).count('1')

    def parse_xor(self, xorclause):
        assert re.search(r'^x( *-?\d+ )*0$', xorclause)

        tmp = xorclause[1:]
        lits = [int(elem) for elem in tmp.split()]
        assert lits[len(lits)-1] == 0

        # remove last element, the 0
        lits = lits[:len(lits)-1]

        return lits

    def xor_to_cnf_simple(self, xorclause, equals=True):
        assert equals is True or equals is False
        if equals is True:
            equals = 1
        else:
            equals = 0

        lits = self.parse_xor(xorclause)

        # empty XOR clause is TRUE, so is NOT an empty clause (i.e. UNSAT)
        if len(lits) == 0:
            return []

        ret = []
        for i in range(2**(len(lits))):
            # only the ones we need
            cls = ""
            if self.popcount(i) % 2 == equals:
                continue

            for at in range(len(lits)):
                if ((i >> at) & 1) == 0:
                    cls += "%d " % lits[at]
                else:
                    cls += "%d " % (-1*lits[at])

            cls += "0"
            ret.append(cls)

        return ret

    def cut_up_xor_to_n(self, xorclause, oldmaxvar):
        assert self.cutsize > 2

        lits = self.parse_xor(xorclause)
        xors = []

        # xor clause that doesn't need to be cut up
        if len(lits) <= self.cutsize:
            retcl = "x"
            for lit in lits:
                retcl += "%d " % lit
            retcl += "0"
            return [[retcl], oldmaxvar]

        at = 0
        newmaxvar = oldmaxvar
        while(at < len(lits)):

            # until when should we cut?
            until = min(at + self.cutsize-1, len(lits))

            # if in the middle, don't add so much
            if at > 0 and until < len(lits):
                until -= 1

            thisxor = "x"
            for i2 in range(at, until):
                thisxor += "%d " % lits[i2]

            # add the extra variables
            if at == 0:
                # beginning, add only one
                thisxor += "%d 0" % (newmaxvar+1)
                newmaxvar += 1
            elif until == len(lits):
                # end, only add the one we already made
                thisxor += "-%d 0" % (newmaxvar)
            else:
                thisxor += "-%d %d 0" % (newmaxvar, newmaxvar+1)
                newmaxvar += 1

            xors.append(thisxor)

            # move along where we are at
            at = until

        return [xors, newmaxvar]

    def num_extra_vars_cls_needed(self, numlits):
        def cls_for_plain_xor(numlits):
            return 2**(numlits-1)

        varsneeded = 0
        clsneeded = 0

        at = 0
        while(at < numlits):
            # at the beginning
            if at == 0:
                if numlits > self.cutsize:
                    at += self.cutsize-1
                    varsneeded += 1
                    clsneeded += cls_for_plain_xor(self.cutsize)
                else:
                    at = numlits
                    clsneeded += cls_for_plain_xor(numlits)

            # in the middle
            elif at + (self.cutsize-1) < numlits:
                at += self.cutsize-2
                varsneeded += 1
                clsneeded += cls_for_plain_xor(self.cutsize)
            # at the end
            else:
                clsneeded += cls_for_plain_xor(numlits-at+1)
                at = numlits

        return [varsneeded, clsneeded]

    def get_stats(self, infilename):
        infile = open(infilename, "r")

        maxvar = 0
        numcls = 0
        extravars_needed = 0
        extracls_needed = 0
        for line in infile:
            line = line.strip()

            # empty line, skip
            if len(line) == 0:
                continue

            # header or comment
            if line[0] == 'p' or line[0] == 'c':
                continue

            # get max var
            maxvar = max(self.get_max_var(line), maxvar)

            if line[0] == 'x':
                e_var, e_clause = self.num_extra_vars_cls_needed(len(self.parse_xor(line)))
                extravars_needed += e_var
                extracls_needed += e_clause
            else:
                numcls += 1

        infile.close()

        return [maxvar, numcls, extravars_needed, extracls_needed]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import optparse
import time
import helper


class Queries (helper.QueryHelper):
    def __init__(self, dbfname):
        super(Queries, self).__init__(dbfname)

    def create_indexes(self):
        helper.drop_idxs(self.c)

        print("Creating needed indexes...")
        t = time.time()
        q = """
        create index `idxclid-del` on `cl_last_in_solver` (`clauseID`, `conflicts`);
        create index `idxclid-del2` on `used_clauses` (`clauseID`);
        create index `idxclid-del3` on `used_clauses` (`clauseID`, `used_at`);
        create index `idxclid1-2` on `clause_stats` (`clauseID`);
        """

        for l in q.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Creating index: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Index creation T: %-3.2f s" % (time.time() - t2))

        print("indexes created T: %-3.2f s" % (time.time() - t))


    def check_only_one(self):
        only_ones = [
            # clause should only be deleted once, since we cleaned this up already
            # in clean_update_data.py
            {"tbl":"cl_last_in_solver", "elem":"clauseID"},
            {"tbl":"clause_stats", "elem":"clauseID"},
            {"tbl":"restart", "elem":"restartID"},
            {"tbl":"restart_dat_for_cl", "elem":"clauseID"},
        ]

        for only_one in only_ones:
            t = time.time()
            q = """
            select c, {elem}
            from (select count() as c, {elem}
                from {tbl}
                group by {elem}
            )
            where c > 1
            """.format(**only_one)
            cursor = self.c.execute(q)
            for row in cursor:
                if row[0] > 1:
                    print("ERROR: More than one of {elem} in {tbl} -- clause ID is: {clid}, count is: {cnt}"
                          .format(**only_one, clid=row[1], cnt=row[0]))
                    exit(-1)
            print("Checked for {tbl} only containing at most one of the same {elem} in {t:2.3f} seconds".
                  format(**only_one, t=(time.time()-t)))

    def check_all_clauses_have_at_most_one_del(self):
        t = time.time()
        q = """
        select f.clid, f.cnt, cl_min, cl_max
        from (
        select cl_last_in_solver.clauseID as clid, count(cl_last_in_solver.clauseID) as cnt, min(conflicts) as cl_min, max(conflicts) as cl_max
        from cl_last_in_solver
        group by clauseID) as f
        where f.cnt != 1
        order by cl_max
        """
        cursor = self.c.execute(q)
        bad = False
        for row in cursor:
            bad = True
            print("ERROR: Clause deleted more than once! ClauseID: {clid}, times deleted: {num}, first deleted: {first}, last deleted: {last}".format(
                clid=row[0], num=row[1], first=row[2], last=row[3]))

        if bad:
            exit(-1)

        print("Checked all clauses have exactly one delete point. T: %-2.3f" % (time.time()-t))

    def check_all_clauses_have_N(self):
        t = time.time()
        Ns = [
            {"tbl1":"reduceDB", "tbl2":"cl_last_in_solver", "elem":"clauseID"},
          ]
        for n in Ns:
            q = """
            select {tbl1}.{elem}
            from {tbl1} left join {tbl2}
            on {tbl1}.{elem}={tbl2}.{elem}
            where {tbl2}.{elem} is NULL
            order by {tbl1}.{elem}
            """.format(**n)
            cursor = self.c.execute(q)
            bad = False
            for row in cursor:
                bad = True
                print("ERROR: {elem}={clid} has {tbl1} but no corresponding {tbl2}:".format(
                    **n, clid=row[0]))

            if bad:
                exit(-1)

        print("Checked all clauses have a %s. T: %-2.3f" % (Ns, time.time()-t))

    def check_glue_sizes(self):
        if options.slow:

            queries = """
            create index `idx-check-qual1` on `reduceDB` (`clauseID`);
            create index `idx-check-qual2` on `clause_stats` ( `clauseID`);
            """
            for l in queries.split('\n'):
                self.c.execute(l)

            t = time.time()
            q = """
            select * from `clause_stats`,`reduceDB`
            where clause_stats.clauseID = reduceDB.clauseID
            and glue > orig_glue"""
            cursor = self.c.execute(q)
            for row in cursor:
                print("ERROR: glue is larger than orig_glue!")
                print(row)
                exit(-1)

            queries = """
            drop index `idx-check-qual1`;
            drop index `idx-check-qual2`;
            """
            for l in queries.split('\n'):
                self.c.execute(l)


            print("Checked for glue vs orig_glue in %-2.3f seconds" % (time.time()-t))

        t = time.time()
        q = """select * from `clause_stats` where orig_glue > glue_before_minim"""
        cursor = self.c.execute(q)
        for row in cursor:
            print("ERROR: orig_glue is larger than glue_before_minim!")
            print(row)
            exit(-1)
        print("Checked for orig_glue vs glue_before_minim in %-2.3f seconds" % (time.time()-t))

        t = time.time()
        q = """select * from `clause_stats` where orig_glue > orig_size"""
        cursor = self.c.execute(q)
        for row in cursor:
            print("ERROR: orig_glue is larger than orig_size!")
            print(row)
            exit(-1)
        print("Checked for orig_glue vs orig_size in %-2.3f seconds" % (time.time()-t))

    def check_is_null(self):

        is_nulls = [
            {"tbl":"restart", "col":"clauseID"},
            {"tbl":"restart_data_for_var", "col":"clauseID"}
        ]

        t = time.time()
        for is_null in is_nulls:
            q = """
            select * from {tbl} where {col} is not NULL
            """.format(**is_null)
            cursor = self.c.execute(q)
            bad = False
            for row in cursor:
                bad = True
                print("ERROR: {col} is not null in table {tbl}: {row}".format(**is_null), row=row)

            if bad:
                exit(-1)

        print("Checked that some things are NULL. T: %-2.3f" % (time.time()-t))

    def check_incorrect_data_values(self):
        incorrect = [
            {"table":"clause_stats", "cond":" orig_glue = 0 and orig_size >= 2"},
            {"table":"clause_stats", "cond":" orig_glue = 1 and orig_size >= 2"},
            {"table":"clause_stats", "cond":" orig_glue > 100000"},
            {"table":"reduceDB", "cond":" glue > 100000"},
            {"table":"clause_stats", "cond":" glue_before_minim = 0 and orig_size >= 2"},
            {"table":"clause_stats", "cond":" glue_before_minim = 1 and orig_size >= 2"},
            {"table":"reduceDB", "cond":" act_ranking < 0"},
        ]
        for incorr in incorrect:
            t = time.time()
            q = """select * from `{table}` where {cond} """.format(**incorr)
            cursor = self.c.execute(q)
            for row in cursor:
                print("ERROR: Following data in `{table}` has {cond}: ".format(**incorr))
                print(row)
                exit(-1)
            print("Checked for %s in %-2.3f seconds" % (q, time.time()-t))

    def check_is_decision_unchanged(self):
        print("Checking if is_decision hasn't changed while solving...")

        q = """
        select clstats.clauseID, rdb.is_decision, clstats.is_decision from
        reduceDB as rdb
        join clause_stats as clstats
        on rdb.clauseID = clstats.clauseID
        where rdb.is_decision != clstats.is_decision
        """
        cursor = self.c.execute(q)
        for row in cursor:
            clid = int(row[0])
            is_dec_rdb = int(row[1])
            is_dec_clstats = int(row[2])
            print("OOps, for clauseID {clid}, RDB's is_decision is {is_dec_rdb}, while clause_stats's is_decision is {is_dec_clstats}".format(clid=clid, is_dec_rdb=is_dec_rdb, is_dec_clstats=is_dec_clstats))
            exit(-1)

        print("Check for is_decision change finished, all good, it never changed")


    def check_at_least_n(self):
        checks = [
            {"table":"clause_stats", "cond":" orig_glue >= 2", "n": 100},
            {"table":"reduceDB", "cond":" dump_no == 0", "n": 10},
            {"table":"reduceDB", "cond":" dump_no == 1", "n": 10}
        ]
        for check in checks:
            q = """select count() from `{table}` where {cond}""".format(**check)
            cursor = self.c.execute(q)
            for row in cursor:
                num = int(row[0])
                print("Number of {cond} in {table}: {num}".format(num=num, **check))
                if num < check["n"]:
                    print("ERROR: That's too few")
                    exit(-1)

            print("Checked for %s" % q)


    def check_non_negative(self):
        tables = ["reduceDB", "clause_stats", "reduceDB_common"]
        for table in tables:
            cols = helper.get_columns(table, options.verbose, self.c)
            for col in cols:
                t = time.time()
                q = """
                select * from `%s` where `%s` < 0
                """ % (table, col)
                cursor = self.c.execute(q)
                for row in cursor:
                    print("ERROR: following data has %s < 0 in table %s: " % (col , table))
                    print(row)
                    exit(-1)
                print("Checked for %s < 0 in table %s. All are >= 0. T: %-3.2f s" %
                      (col, table, time.time() - t))


    def check_positive(self):
        check_zero = [
            ["glue_before_minim", "clause_stats"],
            ["orig_glue", "clause_stats"],
            ["orig_size", "clause_stats"],
            ["size", "reduceDB"],
            ["glue", "reduceDB"],
            ["act_ranking", "reduceDB"], # all ranking starts at 1, not 0
            ["uip1_ranking", "reduceDB"],
            ["prop_ranking", "reduceDB"]
            ]

        for col,table in check_zero:
            t = time.time()
            q = """
            select * from `%s` where `%s` <= 0
            """ % (table, col)
            cursor = self.c.execute(q)
            for row in cursor:
                print("ERROR: Following data from table %s has %s as non-positive: " % (table, col))
                print(row)
                exit(-1)
            print("Checked for %s in table %s. All are positive T: %-3.2f s" %
                  (col, table, time.time() - t))

    def drop_idxs(self):
        helper.drop_idxs(self.c)

if __name__ == "__main__":
    usage = "usage: %prog [options] sqlitedb"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")
    parser.add_option("--slow", action="store_true", default=False,
                      dest="slow", help="Do more checks")

    (options, args) = parser.parse_args()

    if len(args) < 1:
        print("ERROR: You must give the sqlite file!")
        exit(-1)

    with Queries(args[0]) as q:
        #q.create_indexes()
        q.check_all_clauses_have_at_most_one_del()
        q.check_all_clauses_have_N()
        q.check_only_one()
        q.check_glue_sizes()
        q.check_non_negative()
        q.check_positive()
        q.check_incorrect_data_values()
        q.check_at_least_n()
        q.check_is_decision_unchanged()
        #q.drop_idxs()

    print("Done.")



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import unittest

def inv(l):
    assert len(l) > 0
    if l[0] == "-":
        return l[1:]
    else:
        return "-"+l

def check_clause_sanity(c):
    for i1 in range(len(c)):
        l1 = c[i1]
        for i2 in range(i1+1, len(c)):
            l2 = c[i2]
            if l1 == l2:
                print("ERROR: Same literal twice in clause: %s", l1)
                assert False

            if l1 == inv(l2):
                print("ERROR: Inverted literal in clause: %s ", l1)
                assert False

    return True


def can_resolve(c1, c2):
    check_clause_sanity(c1)
    check_clause_sanity(c2)

    # find inverted
    resolve_on = None
    for a in c1:
        for b in c2:
            if a==inv(b):
                if resolve_on is not None:
                    return None
                resolve_on = a

    return resolve_on

def resolve(c1, c2):
    resolve_on = can_resolve(c1,c2)
    assert resolve_on is not None, "Cannot resolve the two clauses!"

    ret = []
    for a in c1:
        if a != resolve_on:
            ret.append(a)

    for b in c2:
        if b != inv(resolve_on) and b not in ret:
            ret.append(b)

    ret = sorted(ret)
    return ret


def ham_w(num):
    w = 0
    for i in range(16):
        w += (num>>i)&1

    return w


def gen_cls(x):
    check_clause_sanity(x)

    ret = []
    for i in range(1<<len(x)):
        if ham_w(i)%2 == (len(x)%2):
            continue

        cl = []
        for i2 in range(len(x)):
            l = x[i2]
            if i>>i2&1 == 1:
                cl.append(inv(l))
            else:
                cl.append(l)

        ret.append(cl)

    return ret


class TestMethods(unittest.TestCase):

    def test_inv(self):
        self.assertEqual(inv("a"), "-a")
        self.assertEqual(inv("-a"), "a")
        self.assertEqual(inv("x1"), "-x1")
        self.assertEqual(inv("-x1"), "x1")

    def test_sanity_good(self):
        self.assertTrue(check_clause_sanity(["a"]))
        self.assertTrue(check_clause_sanity(["a", "b"]))
        self.assertTrue(check_clause_sanity(["a", "-b"]))

    def test_sanity(self):
        with self.assertRaises(AssertionError):
            check_clause_sanity(["a", "a"])
        with self.assertRaises(AssertionError):
            check_clause_sanity(["a", "-a"])

    def tets_gen_cls(self):
        self.assertEqual(gen_cls(["a"]), [["a"]])
        self.assertEqual(gen_cls(["-a"]), [["-a"]])
        self.assertEqual(gen_cls(["a", "b"]), [["a", "b"], ["-a", "-b"]])
        self.assertEqual(gen_cls(["a", "-b"]), [["a", "-b"], ["-a", "b"]])
        self.assertEqual(
            gen_cls(["a", "b", "c"]),
            [["a", "b", "c"], ["-a", "-b", "c"], ["-a", "b", "-c"], ["a", "-b", "-c"]])

    def test_ham_w(self):
        self.assertEqual(ham_w(1), 1)
        self.assertEqual(ham_w(0), 0)
        self.assertEqual(ham_w(2), 1)
        self.assertEqual(ham_w(3), 2)
        self.assertEqual(ham_w(9), 2)
        self.assertEqual(ham_w(8), 1)
        self.assertEqual(ham_w(11), 3)

    def test_resolve(self):
        with self.assertRaises(AssertionError):
            resolve(["-a", "b"], ["-a", "c"])

        with self.assertRaises(AssertionError):
            resolve(["-a", "-b"], ["a", "b"])

        self.assertEqual(resolve(["a", "b", "c"], ["-a", "b"]), ["b", "c"])

def num_inside(cl,lits):
    num = 0
    for l in cl:
        for l2 in lits:
            if l == l2:
                num+=1

    return num


def find_most_inside(xors, to_prove):
    most_inside = 0
    most_inside_cl = None
    for i_x in range(len(xors)):
        x = xors[i_x]
        cls = gen_cls(x)
        for i_c in range(len(cls)):
            cl = cls[i_c]
            n = num_inside(cl, to_prove)
            if n > most_inside:
                most_inside = n
                most_inside_cl = cl
                most_inside_at = i_x

    return most_inside, most_inside_cl, most_inside_at


def find_resolvent_most_inside(xors, cl_other, to_prove):
    most_inside = 0
    most_inside_cl = None
    for i_x in range(len(xors)):
        x = xors[i_x]
        cls = gen_cls(x)
        for i in range(len(cls)):
            cl = cls[i]
            if can_resolve(cl,cl_other) is None:
                continue

            n = num_inside(cl, to_prove)
            if n > most_inside:
                most_inside = n
                most_inside_cl = cl
                most_inside_at = i_x

    return most_inside, most_inside_cl, most_inside_at

def prove(xors, to_prove):
    to_prove = sorted(to_prove)

    ret = []
    most_inside, cl, most_inside_at = find_most_inside(xors, to_prove)
    assert most_inside > 0
    print("cl picked:", cl)
    xors.remove(xors[most_inside_at])

    while cl != to_prove:
        most_inside, cl2, most_inside_at = find_resolvent_most_inside(xors, cl, to_prove)
        assert most_inside > 0
        print("cl picked:", cl2)

        cl = resolve(cl, cl2)
        cl = sorted(cl)
        print("resolvent:", cl)

if __name__ == '__main__':
    # XORs to add together (we know this)
    xors = [
        ["a", "b", "c"],
        ["a", "b", "d"]
    ]

    # so c+d = 0
    # so c=1 d=0 must be banned
    # generate clause
    to_prove = ["-c", "d"]


    print("xors:", xors)
    print("to_prove:", to_prove)
    prove(xors, to_prove)
    print("----------------")
    print("----------------")


    print(gen_cls(["v1", "-v2"]))
    print(gen_cls(["v1", "v2", "v3"]))
    print("----------------")
    print("----------------")

    unittest.main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import os
import optparse
import subprocess
import time
import gzip
import threading
import glob


class PlainHelpFormatter(optparse.IndentedHelpFormatter):

    def format_description(self, description):
        if description:
            return description + "\n"
        else:
            return ""

usage = "usage: %prog [options] cryptominisat5-binary testfile(s)"
desc = """Test solver against some problems
"""

parser = optparse.OptionParser(usage=usage, description=desc,
                               formatter=PlainHelpFormatter())

parser.add_option("--verbose", "-v", action="store_true", default=False,
                  dest="verbose", help="Print more output")
parser.add_option("--threads", "-t", default=4, type=int,
                  dest="threads", help="Number of threads")
parser.add_option("--minisat", type=str,
                  dest="minisat_exe", help="MiniSat location")
parser.add_option("--cms", type=str,
                  dest="cms_exe", help="CryptoMiniSat location")


(options, args) = parser.parse_args()
print_lock = threading.Lock()
todo_lock = threading.Lock()

if len(args) < 1:
    print("ERROR: You must call this script with at least one argument, a file to check")
    exit(-1)


def go_through_cnf(f):
    for line in f:
        line = line.decode('ascii').strip()
        if len(line) == 0:
            continue
        if line[0] == "p":
            line = line.split()
            assert line[1].strip() == "cnf"
            assert line[2].isdigit()
            return int(line[2])

    assert False


def find_num_vars(fname):

    try:
        with gzip.open(fname, 'rb') as f:
            maxvar = go_through_cnf(f)
    except IOError:
        with open(fname, 'rb') as f:
            maxvar = go_through_cnf(f)

    return maxvar

todo = []
exitnum = 0


class MyThread(threading.Thread):
    def __init__(self, threadID, extraopts):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.extraopts = extraopts

    def run(self):
        global todo
        global exitnum
        while len(todo) > 0:
            with todo_lock:
                fname = todo[0]
                todo = todo[1:]

            if options.verbose:
                with print_lock:
                    print("Thread %d pikced up %s" % (self.threadID, fname))

            ret = self.test_velim_one_file(fname)

            with todo_lock:
                exitnum |= ret

        if options.verbose:
            with print_lock:
                print("Finished thread %d" % self.threadID)

    def test_velim_one_file(self, fname):
        orig_num_vars = find_num_vars(fname)

        simp_fname = "simp.out-%d" % self.threadID
        try:
            os.unlink(simp_fname)
        except:
            pass

        toprint = ""

        toexec = [options.cms_exe, "--zero-exit-status",
                  "--preproc", "1", "--verb", "0"]
        toexec.extend(self.extraopts)
        toexec.extend([fname, simp_fname])

        toprint += "Executing: %s\n" % toexec
        with print_lock:
            print(toprint)
        toprint = ""

        start = time.time()
        cms_out_fname = "cms-%s.out" % os.path.split(fname)[1]
        try:
            with open(cms_out_fname, "w") as f:
                subprocess.check_call(" ".join(toexec), stdout=f, shell=True)
        except subprocess.CalledProcessError:
            toprint += "*** ERROR CryptoMiniSat errored out!\n"
            with print_lock:
                print(toprint)
            exit(-1)
            return -1
        t_cms = time.time()-start
        num_vars_after_cms_preproc = find_num_vars(simp_fname)

        start = time.time()
        toexec = [options.minisat_exe, fname]
        toprint += "Executing: %s\n" % toexec
        minisat_out_fname = "minisat_elim_data.out-%d" % self.threadID
        try:
            with open(minisat_out_fname, "w") as f:
                subprocess.check_call(" ".join(toexec), stdout=f, shell=True)
        except subprocess.CalledProcessError:
            toprint += "** Minisat errored out...\n"
            with print_lock:
                print(toprint)
            exit(-1)
            return -1
        t_msat = time.time()-start

        var_elimed = None
        num_vars_after_ms_preproc = None
        with open(minisat_out_fname, "r") as f:
            for line in f:
                line = line.strip()
                if "num-vars-eliminated" in line:
                    var_elimed = int(line.split()[1])
                if "num-free-vars" in line:
                    num_vars_after_ms_preproc = int(line.split()[1])

        assert var_elimed is not None, "Couldn't find var-elimed line, out: %s" % toprint
        assert num_vars_after_ms_preproc is not None, "Couldn't find num-free-vars line, out: %s" % toprint

        toprint += "-> orig num vars: %d\n" % orig_num_vars
        toprint += "-> T-cms : %-4.2f free vars after: %-9d\n" % (t_cms, num_vars_after_cms_preproc)
        toprint += "-> T-msat: %-4.2f free vars after: %-9d\n" % (t_msat, num_vars_after_ms_preproc)
        diff = num_vars_after_cms_preproc - num_vars_after_ms_preproc
        limit = float(orig_num_vars)*0.05
        if diff < limit*8 and t_msat > t_cms*4 and t_msat > 20:
            toprint += " * MiniSat didn't timeout, but we did, acceptable difference.\n"
            with print_lock:
                print(toprint)
            return 0

        if diff > limit:
            toprint += "*** ERROR: No. vars difference %d is more than 5%% " % diff
            toprint += "of original no. of vars, %d\n" % limit
            with print_lock:
                print(toprint)
            return 1

        toprint += "------------------[ thread %d ]------------------------" % self.threadID

        with print_lock:
            print(toprint)

        return 0


def test(extraopts):
    global exitnum
    exitnum = 0
    global todo
    assert os.path.isdir(args[0])
    path = os.path.join(args[0], '')
    todo = glob.glob(path+"/*.cnf.gz")

    threads = []
    for i in range(options.threads):
        threads.append(MyThread(i, extraopts))

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    if exitnum == 0:
        print("ALL PASSED")
    else:
        print("SOME CHECKS FAILED")

    return exitnum

if __name__ == "__main__":
    test(["--preschedule", "occ-bve,must-renumber"])
    if exitnum != 0:
        exit(exitnum)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


import numpy as np
import pandas as pd
import xgboost as xgb
import os
from ccg import *

# to test memory safety
#import gc
# to test memory usage
#from memory_profiler import *

MISSING=np.NaN

raw_data = [
    "is_ternary_resolvent",
    "rdb0.which_red_array",
    "rdb0.last_touched",
    "rdb0.act_ranking_rel",
    "rdb0.uip1_ranking_rel",
    "rdb0.prop_ranking_rel",
    "rdb0.last_touched_diff",
    "cl.time_inside_solver",
    "rdb0.props_made",
    "rdb0_common.avg_props",
    #"rdb0_common.avg_glue", CANNOT DO, ternaries have no glue!
    "rdb0_common.avg_uip1_used",
    "rdb0_common.conflSizeHistLT_avg",
    "rdb0_common.glueHistLT_avg",
    "rdb0.sum_props_made",
    "rdb0.discounted_props_made",
    "rdb0.discounted_props_made2",
    "rdb0.discounted_props_made3",
    "rdb0.discounted_uip1_used",
    "rdb0.discounted_uip1_used2",
    "rdb0.discounted_uip1_used3",
    "rdb0.sum_uip1_used",
    "rdb0.uip1_used",
    "rdb0.size",
    "rdb0.sum_uip1_per_time_ranking",
    "rdb0.sum_props_per_time_ranking",
    "rdb0.sum_uip1_per_time_ranking_rel",
    "rdb0.sum_props_per_time_ranking_rel",
    "rdb0.is_distilled",
    "cl.glueHist_avg",
    "cl.atedecents_binIrred",
    "cl.glueHistLT_avg",
    "cl.glueHist_longterm_avg",
    "cl.num_antecedents",
    "cl.overlapHistLT_avg",
    "cl.conflSizeHist_avg",
    "cl.atedecents_binRed",
    "cl.num_total_lits_antecedents",
    "cl.numResolutionsHistLT_avg",
    "rdb0.glue",
    "cl.orig_glue",
    "cl.glue_before_minim",
    "cl.trail_depth_level",
    #"sum_uip1_per_time_ranking_rel",
    #"sum_props_per_time_ranking_rel",
]

# check reproducibility by dumping and checking against previous run's dump
def dump_or_check(fname, df):
    if check_file_exists(fname):
        df_saved = pd.read_pickle(fname)
        print("Checking equals...", fname)
        if not df.equals(df_saved):
            print("df.shape       :", df.shape)
            print("df_saved.shape :", df_saved.shape)
            assert df.shape == df_saved.shape
            for i in range(df):
                if df[i] != df_saved[i]:
                    print("Not equal:")
                    print(df[i])
                    print(df_saved[i])
                    exit(-1)

    else:
        df.to_pickle(fname)
        print("Not checking, writing: ", fname)


def check_file_exists(fname):
    return os.path.exists(fname)

def get_features(fname):
    best_features = []
    if not check_file_exists(fname):
        print("File '%s' not accessible" % fname)
        exit(-1)

    with open(fname, "r") as f:
        for l in f:
            l = l.strip()
            if len(l) == 0:
                continue

            if l[0] == "#":
                continue

            best_features.append(l)

    return best_features


def check_against_binary_dat(fname, df, df_raw):
    global best_features
    check_file_exists(fname)
    df2 = pd.read_csv(fname, sep=",", names=best_features)
    print("Checking binary vs python:", fname)

    for f in best_features:
        for i in range(df.shape[0]):
            s_pyt = float(df[f][i])
            s_bin = float(df2[f][i])
            if np.isnan(s_pyt) and np.isnan(s_bin):
                continue
            if np.isnan(s_pyt) and not np.isnan(s_bin):
                assert False
            if not np.isnan(s_pyt) and np.isnan(s_bin):
                assert False
            diff = abs(s_pyt - s_bin)

            if diff > 10e-5:
                for f_raw in list(df_raw):
                    print("pyt_raw:", df_raw[f_raw][i], " feat: ", f_raw)
                print("pyt:", s_pyt, " feat: ", f)
                print("bin:", s_bin, " feat: ", f)
                print("diff for feat ", f, ": ", diff)
            else:
                #print("OK for feat", f)
                pass
            assert diff < 10e-5

    #assert df.equals(df2)


models = []
best_features = []
feat_gen_exprs = []
feat_gen_funcs = []

def add_features(df, df2):
    for i, feat_gen_func in zip(range(len(best_features)), feat_gen_funcs):
        df2[:, i] = feat_gen_func(df)

def set_up_features(features_fname):
    global best_features
    global feat_gen_exprs
    global feat_gen_funcs
    best_features = get_features(features_fname)
    for i, feat in zip(range(len(best_features)), best_features):
        feat_gen_expr = ccg.to_source(ast.parse(feat))
        feat_gen_exprs.append(feat_gen_expr)
        create_function = "def a%d(df): return %s" % (i, feat_gen_expr)
        exec(create_function)
        exec("feat_gen_funcs.append(a%d)" % i)
    #print(feat_gen_funcs)


def load_models(short_fname, long_fname, forever_fname):
    global models
    for fname in [short_fname, long_fname, forever_fname]:
        clf_xgboost = xgb.XGBRegressor(n_jobs=1)
        new_fname = fname.replace("-py.", "-xgb.")
        if not os.path.exists(new_fname):
            new_fname = fname.replace("-py.", ".")
        clf_xgboost.load_model(new_fname)
        models.append(clf_xgboost)


num_called = 0

# to test memory usage
#@profile
def predict(data, check=False, dump=False):
    global num_called
    ret = []

    df = pd.DataFrame(data, columns=raw_data)
    transformed_data = np.empty((df.shape[0], len(best_features)), dtype=float)
    if check:
        dump_or_check('df_check'+str(num_called), df)

    add_features(df, transformed_data)
    df_final = pd.DataFrame(transformed_data, columns=best_features)
    df_final.replace([np.inf, np.NaN, np.inf, np.NINF, np.Infinity], MISSING, inplace=True)

    if check:
        check_against_binary_dat('bin_dump'+str(num_called)+".csv", df_final, df)

    for i in range(3):
        #x = models[i].predict(df_final)
        x = models[i].get_booster().inplace_predict(df_final)
        if check:
            dump_or_check('x-%d-%d' % (num_called, i), pd.DataFrame(x))
        ret.append(x)

    if dump:
        for i,name in zip(range(3), ["short", "long", "forever"]):
            df["x.used_later_%s" % name] = ret[i]

        df.to_pickle('df_dump_%s' % num_called)

    if check:
        dump_or_check('df_pred'+str(num_called), df_final)

    num_called += 1
    del df
    del df_final
    del transformed_data
    # gc.collect() # to test memory safety
    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import with_statement  # Required in 2.5
from __future__ import print_function
import random

def shuffle_cnf(fname1, fname2, seed):
    random.seed(int(seed))

    with open(fname1, "r") as f1:
        headers = []
        lines = []
        for line in f1:
            line = line.strip()
            if len(line) > 1 and line[0] == 'p':
                headers.append(line)
            else:
                lines.append(line)

    random.shuffle(lines)

    with open(fname2, "w") as f2:
        for line in headers:
            f2.write(line+"\n")
        for line in lines:
            f2.write(line+"\n")

def get_max_var_from_clause(line):
    maxvar = 0
    # strip leading 'x'
    line2 = line.strip()
    if len(line2) > 0 and line2[0] == 'x':
        line2 = line2[1:]

    for lit in line2.split():
        num = 0
        try:
            num = int(lit)
        except ValueError:
            print("line '%s' contains a non-integer variable" % line2)

        maxvar = max(maxvar, abs(num))

    return maxvar


class debuglib:
    @staticmethod
    def generate_random_assumps(maxvar):
        assumps = ""
        num = 0
        varsInside = set()

        # Half of the time, no assumptions at all
        if random.randint(0, 1) == 1:
            return assumps

        # use a distribution so that few will be in assumps
        while (num < maxvar and random.randint(0, 4) > 0):

            # get a var that is not already inside the assumps
            thisVar = random.randint(1, maxvar)
            tries = 0
            while (thisVar in varsInside):
                thisVar = random.randint(1, maxvar)
                tries += 1

                # too many tries, don't waste time
                if tries > 100:
                    return assumps

            varsInside.add(thisVar)

            # random sign
            num += 1
            if random.randint(0, 1):
                thisVar *= -1

            assumps += "%d " % thisVar

        return assumps

    @staticmethod
    def file_len_no_comment(fname):
        i = 0
        with open(fname) as f:
            for l in f:
                # ignore comments and empty lines and header
                if not l or l[0] == "c" or l[0] == "p":
                    continue
                i += 1

        return i

    @staticmethod
    def main(fname1, fname2):

        # approx number of solve()-s to add
        if random.randint(0, 1) == 1:
            num_to_add = random.randint(0, 10)
        else:
            num_to_add = 0

        # based on length and number of solve()-s to add, intersperse
        # file with ::solve()
        file_len = debuglib.file_len_no_comment(fname1)
        if num_to_add > 0:
            nextToAdd = random.randint(1, int(file_len / num_to_add) + 1)
        else:
            nextToAdd = file_len + 1

        fin = open(fname1, "r")
        fout = open(fname2, "w")
        at = 0
        maxvar = 0
        for line in fin:
            line = line.strip()

            # ignore comments (but write them out)
            if not line or line[0] == "c" or line[0] == 'p':
                fout.write(line + '\n')
                continue

            at += 1
            if at >= nextToAdd:
                assumps = debuglib.generate_random_assumps(maxvar)
                if random.choice([True, False]):
                    fout.write("c Solver::solve( %s )\n" % assumps)
                elif random.choice([True, False]):
                    fout.write("c Solver::simplify( %s )\n" % assumps)
                else:
                    fout.write("c Solver::simplify( %s )\n" % assumps)
                    fout.write("c Solver::solve( %s )\n" % assumps)

                nextToAdd = at + \
                    random.randint(1, int(file_len / num_to_add) + 1)

            # calculate max variable
            maxvar = max(maxvar, get_max_var_from_clause(line))

            # copy line over
            fout.write(line + '\n')
        fout.close()
        fin.close()


def intersperse(fname1, fname2, seed):
    random.seed(int(seed))
    debuglib.main(fname1, fname2)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

# pylint: disable=invalid-name,line-too-long,too-many-locals,consider-using-sys-exit

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import sklearn
import sklearn.metrics
import re
import ast
import math
import time
import os.path
import sqlite3
import functools
from ccg import *

try:
    from termcolor import cprint
except ImportError:
    termcolor_avail = False
else:
    termcolor_avail = True

from pprint import pprint
try:
    import mlflow
except ImportError:
    mlflow_avail = False
else:
    mlflow_avail = True


class QueryHelper:
    def __init__(self, dbfname):
        if not os.path.isfile(dbfname):
            print("ERROR: Database file '%s' does not exist" % dbfname)
            exit(-1)

        self.conn = sqlite3.connect(dbfname)
        self.c = self.conn.cursor()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.commit()
        self.conn.close()


class QueryFill (QueryHelper):
    def create_indexes(self, verbose=False, used_clauses="used_clauses"):
        t = time.time()
        print("Recreating indexes...")
        queries = """
        create index `idxclid6-4` on `reduceDB` (`clauseID`, `conflicts`)
        create index `idxclidUCLS-2` on `{used_clauses}` ( `clauseID`, `used_at`);
        create index `idxcl_last_in_solver-1` on `cl_last_in_solver` ( `clauseID`, `conflicts`);
        """.format(used_clauses=used_clauses)
        for l in queries.split('\n'):
            t2 = time.time()

            if verbose:
                print("Creating index: ", l)
            self.c.execute(l)
            if verbose:
                print("Index creation T: %-3.2f s" % (time.time() - t2))

        print("indexes created T: %-3.2f s" % (time.time() - t))

    def delete_and_create_used_laters(self):
        tiers = ["short", "long", "forever"]
        tables = ["used_later", "used_later_anc"]
        for tier in tiers:
            for table in tables:
                q = """
                    DROP TABLE IF EXISTS `{table}_{tier}`;
                """
                self.c.execute(q.format(tier=tier, table=table))

        # Create and fill used_later_X tables
        q_create = """
        create table `{table}_{tier}` (
            `clauseID` bigint(20) NOT NULL,
            `rdb0conflicts` bigint(20) NOT NULL,
            `used_later` float,
            `percentile_fit` float DEFAULT NULL
        );"""
        # NOTE: "percentile_fit" is the top percentile this use belongs to. Filled in later.

        for tier in tiers:
            for table in tables:
                self.c.execute(q_create.format(tier=tier, table=table))

        idxs = """
        create index `{table}_{tier}_idx3` on `{table}_{tier}` (`used_later`);
        create index `{table}_{tier}_idx1` on `{table}_{tier}` (`clauseID`, `rdb0conflicts`);
        create index `{table}_{tier}_idx2` on `{table}_{tier}` (`clauseID`, `rdb0conflicts`, `used_later`);"""

        t = time.time()
        for tier in tiers:
            for table in tables:
                for l in idxs.format(tier=tier, table=table).split('\n'):
                    self.c.execute(l)

        print("used_later* dropped and recreated T: %-3.2f s" % (time.time() - t))

    # The most expesive operation of all, when called with "forever"
    def fill_used_later_X(self, tier, duration, used_clauses="used_clauses",
                          table="used_later"):

        min_del_distance = duration
        if min_del_distance > 2*1000*1000:
            min_del_distance = 100*1000

        # Note: "weight" below is because a child has less weight than a parent
        #       discount is 0.5 (as per fix_up_frat), so a child is 0.5, a grand-child
        #       is 0.25 etc. Below 0.05 we don't care and it's a 0.
        q_fill = """
        insert into {table}_{tier}
        (
        `clauseID`,
        `rdb0conflicts`,
        `used_later`
        )
        SELECT
        rdb0.clauseID
        , rdb0.conflicts
        , sum(ucl.weight) as `used_later`

        FROM
        reduceDB as rdb0
        left join {used_clauses} as ucl

        -- reduceDB is always present, {used_clauses} may not be, hence left join
        on (ucl.clauseID = rdb0.clauseID
            and ucl.used_at > (rdb0.conflicts)
            and ucl.used_at <= (rdb0.conflicts+{duration}))

        join cl_last_in_solver
        on cl_last_in_solver.clauseID = rdb0.clauseID

        WHERE
        rdb0.clauseID != 0
        and cl_last_in_solver.conflicts >= (rdb0.conflicts + {min_del_distance})

        group by rdb0.clauseID, rdb0.conflicts;"""

        t = time.time()
        q = q_fill.format(
            tier=tier, used_clauses=used_clauses,
            duration=duration,
            table=table,
            min_del_distance=min_del_distance)
        self.c.execute(q)

        q_fix_null = "update {table}_{tier} set used_later = 0 where used_later is NULL".format(
            tier=tier, table=table)
        self.c.execute(q_fix_null)


        q_num = "select count(*) from {table}_{tier}".format(tier=tier, table=table)
        self.c.execute(q_num)
        rows = self.c.fetchall()
        for row in rows:
            num = row[0]

        if table == "used_later" and num == 0:
            print("ERROR: number of rows in {table}_{tier} is 0!".format(tier=tier, table=table))
            print("Query was: %s" % q)
            exit(-1)


        print("%s_%s filled T: %-3.2f s -- num rows: %d" %
              (table, tier, time.time() - t, num))

    def fill_used_later_X_perc_fit(self, tier, table):
        print("Filling percentile_fit for {table}_{tier}".format(tier=tier, table=table))

        q = """
        update {table}_{tier}
        set percentile_fit = (
            select max({table}_percentiles.percentile)
            from {table}_percentiles
            where
            {table}_percentiles.type_of_dat="{tier}"
            and {table}_percentiles.percentile_descr="top_non_zero"
            and {table}_percentiles.val >= {table}_{tier}.used_later);
        """
        t = time.time()
        self.c.execute(q.format(tier=tier, table=table))
        print("used_later_%s percentile filled T: %-3.2f s" % (tier, time.time() - t))



def write_mit_header(f):
    f.write("""/******************************************
Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
***********************************************/\n\n""")

def parse_configs(confs):
    match = re.match(r"^([0-9]*)-([0-9]*)$", confs)
    if not match:
        print("ERROR: we cannot parse your config options: '%s'" % confs)
        exit(-1)

    conf_from = int(match.group(1))
    conf_to = int(match.group(2))+1
    if conf_to <= conf_from:
        print("ERROR: Conf range is not increasing")
        exit(-1)

    print("Running configs:", range(conf_from, conf_to))
    return conf_from, conf_to


def get_features(fname):
    best_features = []
    check_file_exists(fname)
    with open(fname, "r") as f:
        for l in f:
            l = l.strip()
            if len(l) == 0:
                continue

            if l[0] == "#":
                continue

            best_features.append(l)

    return best_features


def helper_divide(dividend, divisor, df, features, verb, name=None):
    """
    to be used like:
    import functools
    divide = functools.partial(helper.divide, df=df, features=features, verb=options.verbose)
    """

    # dividend feature not present
    #if dividend not in features:
        #return None

    # divisorfeature not present
    #if divisor not in features:
        #return None

    # divide
    if verb:
        print("Dividing. dividend: '%s' divisor: '%s' " % (dividend, divisor))

    if name is None:
        name = "(%s/%s)" % (dividend, divisor)

    df[name] = df[dividend].div(df[divisor])
    return name

def helper_larger_than(lhs, rhs, df, features, verb):
    """
    to be used like:
    import functools
    larger_than = functools.partial(helper.larger_than, df=df, features=features, verb=options.verbose)
    """

    # divide
    if verb:
        print("Calulating '%s' >: '%s' " % (lhs, rhs))

    name = "(" + lhs + ">" + rhs + ")"
    df[name] = (df[lhs] > df[rhs]).astype(int)
    return name

def helper_add(toadd, df, features, verb):
    """
    to be used like:
    import functools
    larger_than = functools.partial(helper.larger_than, df=df, features=features, verb=options.verbose)
    """

    # add
    if verb:
        print("Calulating: the feature addition of: %s", toadd)

    name = "("
    for i in range(1, len(toadd)):
        name = toadd[i]
        if i < len(toadd)-1:
            name+="+"
    name += ")"

    df[name] = df[toadd[0]]
    for i in range(1, len(toadd)):
        df[name] += df[toadd[i]]
    return name


def dangerous(conn):
    conn.execute("PRAGMA journal_mode = MEMORY")
    conn.execute("PRAGMA synchronous = OFF")


def drop_idxs(conn):
    q = """
    SELECT name FROM sqlite_master WHERE type == 'index'
    """
    conn.execute(q)
    rows = conn.fetchall()
    queries = ""
    for row in rows:
        #print("Will delete index:", row[0])
        queries += "drop index if exists `%s`;\n" % row[0]

    t = time.time()
    for q in queries.split("\n"):
        conn.execute(q)

    print("Removed indexes: T: %-3.2f s"% (time.time() - t))


def get_columns(tablename, verbose, conn):
    q = "pragma table_info(%s);" % tablename
    conn.execute(q)
    rows = conn.fetchall()
    columns = []
    for row in rows:
        if verbose:
            print("Using column in table {tablename}: {col}".format(
                tablename=tablename, col=row[1]))
        columns.append(row[1])

    return columns

def query_fragment(tablename, not_cols, short_name, verbose, conn):
    cols = get_columns(tablename, verbose, conn)
    filtered_cols = list(set(cols).difference(not_cols))
    ret = ""
    for col in filtered_cols:
        ret += ", {short_name}.`{col}` as `{short_name}.{col}`\n".format(
            col=col, short_name=short_name)

    if verbose:
        print("query for short name {short_name}: {ret}".format(
            short_name=short_name, ret=ret))

    return ret


def not_inside(not_these, inside_here):
    for not_this in not_these:
        if not_this in inside_here:
            return False

    return True


# to check for too large or NaN values:
def check_too_large_or_nan_values(df, features=None):
    print("Checking for too large or NaN values...")
    if features is None:
        features = df.columns.values.flatten().tolist()

    index = 0
    for index, row in df[features].iterrows():
        print("-------------")
        print("At row index: ", index)
        for val, name in zip(row, features):
            print("Name: '%s', val: %s" % (name, val))
            if type(val) == str:
                continue

            if math.isnan(val) or not np.isfinite(val) or val > np.finfo(np.float32).max:
                print("issue with feature '%s' Value: '%s'  Type: '%s" % (
                    name, val, type(val)))
        index += 1

    print("Checking finished.")


def print_confusion_matrix(cm,
                           normalize=False,
                           title='Confusion matrix'):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    print(title)
    if mlflow_avail:
        mlflow.log_metric(title, cm[0][0])
    np.set_printoptions(precision=2)
    print(cm)


def calc_min_split_point(df, min_samples_split):
    split_point = int(float(df.shape[0])*min_samples_split)
    if split_point < 10:
        split_point = 10
    print("Minimum split point: ", split_point)
    return split_point


def error_format(error):
    if error is None:
        return "XXX"
    else:
        return "{0:<2.2E}".format(error)


def calc_regression_error(data, features, to_predict, clf, toprint,
                  average="binary", highlight=False):
    X_data = data[features]
    y_data = data[to_predict]
    print("Number of elements:", X_data.shape)
    if data.shape[0] <= 1:
        print("Cannot calculate regression error, too few elements")
        return None
    y_pred = clf.predict(X_data)
    main_error = sklearn.metrics.mean_squared_error(y_data, y_pred)
    print("Mean squared error is: %9s" % error_format(main_error))
    median_absolute_error = sklearn.metrics.median_absolute_error(y_data, y_pred)
    print("Median abs error is  : %9s" % error_format(median_absolute_error))

    # use distrib
    for start,end in [(0,10), (1,10), (10, 100), (100, 1000), (1000,10000), (10000, 1000000)]:
        x = "--> Strata  %6d <= %21s < %8d " % (start, to_predict, end)
        myfilt = data[(data[to_predict] >= start) & (data[to_predict] < end)]
        X_data = myfilt[features]
        y_data = myfilt[to_predict]
        y = " -- elems: {:12}".format(str(X_data.shape))
        if myfilt.shape[0] <= 1:
            msqe = None
            med_abs_err = None
            mean_err = None
        else:
            y_pred = clf.predict(X_data)
            msqe = sklearn.metrics.mean_squared_error(y_data, y_pred)
            med_abs_err = sklearn.metrics.median_absolute_error(y_data, y_pred)
            mean_err = (y_data - y_pred).sum()/len(y_data)
        print("{} {}   msqe: {:9s}   mabse: {:9s} abs: {:9s}".format(
            x, y,  error_format(msqe), error_format(med_abs_err), error_format(mean_err)))

    # glue distrib
    for start,end in [(0,3), (3,8), (8, 15), (15, 25), (25,50), (50, 100), (100, 1000000)]:
        x = "--> Strata  %6d <= %21s < %8d " % (start, "rdb0.glue", end)
        myfilt = data[(data["rdb0.glue"] >= start) & (data["rdb0.glue"] < end)]
        X_data = myfilt[features]
        y_data = myfilt[to_predict]
        y = " -- elems: {:12}".format(str(X_data.shape))
        if myfilt.shape[0] <= 1:
            msqe = None
            med_abs_err = None
            mean_err = None
        else:
            y_pred = clf.predict(X_data)
            msqe = sklearn.metrics.mean_squared_error(y_data, y_pred)
            med_abs_err = sklearn.metrics.median_absolute_error(y_data, y_pred)
            mean_err = (y_data - y_pred).sum()/len(y_data)
        print("{} {}   msqe: {:9s}   mabse: {:9s} abs: {:9s}".format(
            x, y,  error_format(msqe), error_format(med_abs_err), error_format(mean_err)))

    return main_error


def conf_matrixes(data, features, to_predict, clf, toprint,
                  average="binary", highlight=False):
    # get data
    X_data = data[features]
    y_data = data[to_predict]
    print("Number of elements:", X_data.shape)
    if data.shape[0] <= 1:
        print("Cannot calculate confusion matrix, too few elements")
        return None, None, None, None

    # Preform prediction
    def f(x):
        if x > 0.5:
            return 1
        else:
            return 0

    y_pred = clf.predict(X_data)
    #print("type(y_pred[0]): ",  type(y_pred[0]))
    if type(y_pred[0]) == np.float32:
        y_pred = np.array([f(x) for x in y_pred])

    # calc acc, precision, recall
    accuracy = sklearn.metrics.accuracy_score(
        y_data, y_pred)

    precision = sklearn.metrics.precision_score(
        y_data, y_pred, pos_label=1, average=average)


    recall = sklearn.metrics.recall_score(
        y_data, y_pred, pos_label=1, average=average)

    # ROC AUC
    predsi = np.array(y_pred)
    y_testi = pd.DataFrame(y_data)["x.class"].squeeze()
    try:
        roc_auc = sklearn.metrics.roc_auc_score(y_testi, predsi)
    except:
        print("NOTE: ROC AUC is set to 0 because of completely one-sided OK/BAD")
        roc_auc = 0

    # record to mlflow
    if mlflow_avail:
        mlflow.log_metric(toprint + " -- accuracy", accuracy)
        mlflow.log_metric(toprint + " -- precision", precision)
        mlflow.log_metric(toprint + " -- recall", recall)
        mlflow.log_metric(toprint + " -- roc_auc", roc_auc)

    color = "white"
    bckgrnd = "on_grey"
    if highlight:
        color="green"
        bckgrnd = "on_grey"

    txt = "%s prec : %-3.4f  recall: %-3.4f accuracy: %-3.4f roc_auc: %-3.4f"
    vals = (toprint, precision, recall, accuracy, roc_auc)
    if termcolor_avail:
        cprint(txt % vals , color, bckgrnd)
    else:
        cprint(txt % vals)

    # Plot confusion matrix
    cnf_matrix = sklearn.metrics.confusion_matrix(
        y_true=y_data, y_pred=y_pred)
    print_confusion_matrix(
        cnf_matrix,
        title='Confusion matrix without normalization -- %s' % toprint)
    print_confusion_matrix(
        cnf_matrix, normalize=True,
        title='Normalized confusion matrix -- %s' % toprint)

    return roc_auc


def check_file_exists(fname):
    try:
        f = open(fname)
    except IOError:
        print("File '%s' not accessible" % fname)
        exit(-1)
    finally:
        f.close()


def output_to_classical_dot(clf, features, fname):

    feat_tmp = []
    for f in features:
        x = str(f)
        x = x.replace("rdb0.", "")
        x = x.replace("cl.", "")
        x = x.replace("HistLT.", "History_Long_Term")
        x = x.replace("rdb0_common.", "all_learnts")
        feat_tmp.append(x)

    sklearn.tree.export_graphviz(clf, out_file=fname,
                                 feature_names=feat_tmp,
                                 #class_names=clf.classes_,
                                 filled=True, rounded=True,
                                 special_characters=True,
                                 proportion=True)
    print("Run dot:")
    print("dot -Tpng {fname} -o {fname}.png".format(fname=fname))
    print("gwenview {fname}.png".format(fname=fname))


def print_feature_ranking(clf, X_train, top_num_features, features, plot=False):
    best_features = []
    importances = clf.feature_importances_
    std = np.std(
        [tree.feature_importances_ for tree in clf.estimators_], axis=0)
    indices = np.argsort(importances)[::-1]
    indices = indices[:top_num_features]
    myrange = min(X_train.shape[1], top_num_features)

    # Print the feature ranking
    print("Feature ranking:")

    for f in range(myrange):
        print("%-3d  %-55s -- %8.4f" %
              (f + 1, features[indices[f]], importances[indices[f]]))
        best_features.append(features[indices[f]])

    # Plot the feature importances of the clf
    if plot:
        plot_feature_importances(importances, indices, myrange, std, features)

    return best_features


def plot_feature_importances(importances, indices, myrange, std, features):
        plt.figure()
        plt.title("Feature importances")
        plt.bar(range(myrange), importances[indices],
                color="r", align="center",
                yerr=std[indices])
        plt.xticks(range(myrange), [features[x]
                                    for x in indices], rotation=45)
        plt.xlim([-1, myrange])


def add_features_from_fname(df, features_fname, verbose=False):
    print("Adding features...")
    if not os.path.exists(features_fname):
        print("ERROR: Feature file '%s' does not exist" % features_fname)
        exit(-1)

    cldata_add_minimum_computed_features(df, verbose)
    best_features = get_features(features_fname)
    for feat in best_features:
        toeval = ccg.to_source(ast.parse(feat))
        print("Adding feature %s as eval %s" % (feat, toeval))
        df[feat] = eval(toeval)


def add_features_from_list(df, best_features, verbose=False):
    print("Adding features...")
    cldata_add_minimum_computed_features(df, verbose)
    for feat in best_features:
        toeval = ccg.to_source(ast.parse(feat))
        print("Adding feature %s as eval %s" % (feat, toeval))
        df[feat] = eval(toeval)


def make_missing_into_nan(df):
    print("Making None into NaN...")
    def make_none_into_nan(x):
        if x is None:
            return np.nan
        else:
            return x

    for col in list(df):
        if type(None) in df[col].apply(type).unique():
            df[col] = df[col].apply(make_none_into_nan)
    print("Done.")

def cldata_add_minimum_computed_features(df, verbose):
    divide = functools.partial(helper_divide, df=df, features=list(df), verb=verbose)
    divide("rdb0.act_ranking", "rdb0_common.tot_cls_in_db", name="rdb0.act_ranking_rel")
    divide("rdb0.prop_ranking", "rdb0_common.tot_cls_in_db", name="rdb0.prop_ranking_rel")
    divide("rdb0.uip1_ranking", "rdb0_common.tot_cls_in_db", name="rdb0.uip1_ranking_rel")
    divide("rdb0.sum_uip1_per_time_ranking", "rdb0_common.tot_cls_in_db",
           name="rdb0.sum_uip1_per_time_ranking_rel")
    divide("rdb0.sum_props_per_time_ranking", "rdb0_common.tot_cls_in_db",
           name="rdb0.sum_props_per_time_ranking_rel")

    df["rdb0_common.tot_irred_cls"] = df["rdb0_common.num_bin_irred_cls"] + df["rdb0_common.num_long_irred_cls"]
    divide("rdb0_common.tot_irred_cls", "rdb0_common.num_vars")
    divide("rdb0_common.num_long_irred_cls", "rdb0_common.num_long_irred_cls_lits")
    divide("rdb0_common.num_long_irred_cls_lits", "rdb0_common.num_vars")
    divide("rdb0_common.num_long_irred_cls", "rdb0_common.num_vars")


def cldata_add_computed_features(df, verbose):
    print("Adding computed features...")
    cldata_add_minimum_computed_features(df, verbose)

    del df["cl.conflicts"]
    del df["cl.restartID"]
    del df["rdb0.introduced_at_conflict"]

    divide = functools.partial(helper_divide, df=df, features=list(df), verb=verbose)
    larger_than = functools.partial(helper_larger_than, df=df, features=list(df), verb=verbose)
    add = functools.partial(helper_add, df=df, features=list(df), verb=verbose)

    # ************
    # TODO decision level and branch depth are the same, right???
    # ************
    print("size/glue/trail rel...")
    divide("cl.trail_depth_level", "cl.trailDepthHistLT_avg")
    divide("cl.trail_depth_level", "cl.trailDepthHist_avg")

    divide("cl.num_total_lits_antecedents", "cl.num_antecedents")

    del df["rdb0.uip1_ranking"]
    del df["rdb0.prop_ranking"]
    del df["rdb0.act_ranking"]
    del df["rdb0.sum_uip1_per_time_ranking"]
    del df["rdb0.sum_props_per_time_ranking"]
    del df["rdb0_common.tot_cls_in_db"]

    # divide by avg and median
    divide("rdb0.uip1_used", "rdb0_common.avg_uip1_used")
    divide("rdb0.props_made", "rdb0_common.avg_props")
    divide("rdb0.glue", "rdb0_common.avg_glue")

    divide("rdb0.uip1_used", "rdb0_common.median_uip1_used")
    divide("rdb0.props_made", "rdb0_common.median_props")

    time_in_solver = "cl.time_inside_solver"
    sum_props_per_time = divide("rdb0.sum_props_made", time_in_solver)
    sum_uip1_per_time = divide("rdb0.sum_uip1_used", time_in_solver)
    divide(sum_props_per_time, "rdb0_common.median_sum_uip1_per_time")
    divide(sum_uip1_per_time, "rdb0_common.median_sum_props_per_time")
    divide(sum_props_per_time, "rdb0_common.avg_sum_uip1_per_time")
    divide(sum_uip1_per_time, "rdb0_common.avg_sum_props_per_time")
    #del df[time_in_solver]

    divisors = [
        "cl.conflSizeHistlt_avg"
        , "cl.glueHistLT_avg"
        , "rdb0.glue"
        , "rdb0.size"
        # , "cl.orig_connects_num_communities"
        # , "rdb0.connects_num_communities"
        , "cl.orig_glue"
        , "cl.glue_before_minim"
        , "cl.glueHist_avg"
        , "cl.glueHist_longterm_avg"
        # , "cl.decision_level_hist"
        , "cl.numResolutionsHistLT_avg"
        , "cl.trailDepthHistLT_avg"
        , "cl.trailDepthHist_avg"
        , "cl.branchDepthHistQueue_avg"
        , "cl.overlapHistLT_avg"
        , "(cl.num_total_lits_antecedents/cl.num_antecedents)"
        , "cl.num_antecedents"
        , "rdb0.act_ranking_rel"
        , "rdb0.prop_ranking_rel"
        , "rdb0.uip1_ranking_rel"
        , "rdb0.sum_uip1_per_time_ranking_rel"
        , "rdb0.sum_props_per_time_ranking_rel"
        , "cl.time_inside_solver"
        # , "cl.num_overlap_literals"
        ]

    # discounted stuff
    divide("rdb0.discounted_uip1_used", "rdb0_common.avg_uip1_used")
    divide("rdb0.discounted_props_made", "rdb0_common.avg_props")
    divide("rdb0.discounted_uip1_used", "rdb0_common.median_uip1_used")
    divide("rdb0.discounted_props_made", "rdb0_common.median_props")
    #==
    divide("rdb0.discounted_uip1_used2", "rdb0_common.avg_uip1_used")
    divide("rdb0.discounted_props_made2", "rdb0_common.avg_props")
    divide("rdb0.discounted_uip1_used2", "rdb0_common.median_uip1_used")
    divide("rdb0.discounted_props_made2", "rdb0_common.median_props")

    sum_uip1_per_time = divide("rdb0.sum_uip1_used", "cl.time_inside_solver")
    sum_props_per_time = divide("rdb0.sum_props_made", "cl.time_inside_solver")
    antec_rel = divide("cl.num_total_lits_antecedents", "cl.antec_data_sum_sizeHistLT_avg")
    divisors.append(sum_uip1_per_time)
    divisors.append(sum_props_per_time)
    divisors.append("rdb0.discounted_uip1_used")
    divisors.append("rdb0.discounted_props_made")
    divisors.append(antec_rel)

    orig_cols = list(df)

    # Thanks to Chai Kian Ming Adam for the idea of using LOG instead of SQRT
    # add LOG
    if False:
        toadd = []
        for divisor in divisors:
            x = "log2("+divisor+")"
            df[x] = df[divisor].apply(np.log2)
            toadd.append(x)
        divisors.extend(toadd)

    # relative data
    cols = list(df)
    for col in cols:
        if ("rdb" in col or "cl." in col) and "restart_type" not in col and "tot_cls_in" not in col:
            for divisor in divisors:
                divide(divisor, col)
                divide(col, divisor)

    # smaller/larger than
    print("smaller-or-greater comparisons...")
    if False:
        for col in cols:
            if "avg" in col or "median" in col:
                for divisor in divisors:
                    larger_than(col, divisor)

    # smaller-or-greater comparisons
    #if not short:
        #larger_than("cl.antec_data_sum_sizeHistLT_avg", "cl.num_total_lits_antecedents")
        #larger_than("cl.overlapHistLT_avg", "cl.num_overlap_literals")

    # print("flatten/list...")
    #old = set(df.columns.values.flatten().tolist())
    #df = df.dropna(how="all")
    #new = set(df.columns.values.flatten().tolist())
    #if len(old - new) > 0:
        #print("ERROR: a NaN number turned up")
        #print("columns: ", (old - new))
        #assert(False)
        #exit(-1)

def print_datatypes(df):
    pd.set_option('display.max_rows', len(df.dtypes))
    print(df.dtypes)
    pd.reset_option('display.max_rows')



# -*- coding: utf-8 -*-
#
# CryptoMiniSat
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import unicode_literals
from __future__ import print_function
from array import array as _array
import os
import sys
import unittest
import time


import pycryptosat
from pycryptosat import Solver

_MODULE_DIR = os.path.dirname(os.path.realpath(__file__))+os.path.sep

def array(typecode, initializer=()):
    return _array(str(typecode), initializer)


def check_clause(clause, solution):
    for lit in clause:
        var = abs(lit)
        if lit < 0:
            inverted = True
        else:
            inverted = False

        if solution[var] != inverted:
            return True


def check_solution(clauses, solution):
    for clause in clauses:
        if check_clause(clause, solution) is False:
            return False

    return True

# -------------------------- test clauses --------------------------------

# p cnf 5 3
# 1 -5 4 0
# -1 5 3 4 0
# -3 -4 0
clauses1 = [[1, -5, 4], [-1, 5, 3, 4], [-3, -4]]

# p cnf 2 2
# -1 0
# 1 0
clauses2 = [[-1], [1]]

# p cnf 2 3
# -1 2 0
# -1 -2 0
# 1 -2 0
clauses3 = [[-1, 2], [-1, -2], [1, -2]]

# -------------------------- actual unit tests ---------------------------


class TestXor(unittest.TestCase):

    def setUp(self):
        self.solver = Solver(threads=2)

    def test_wrong_args(self):
        self.assertRaises(TypeError, self.solver.add_xor_clause, [1, 2])
        self.assertRaises(ValueError, self.solver.add_xor_clause, [1, 0], True)
        self.assertRaises(
            ValueError, self.solver.add_xor_clause, [-1, 2], True)

    def test_binary(self):
        self.solver.add_xor_clause([1, 2], False)
        res, solution = self.solver.solve([1])
        self.assertEqual(res, True)
        self.assertEqual(solution, (None, True, True))

    def test_unit(self):
        self.solver.add_xor_clause([1], False)
        res, solution = self.solver.solve()
        self.assertEqual(res, True)
        self.assertEqual(solution, (None, False))

    def test_unit2(self):
        self.solver.add_xor_clause([1], True)
        res, solution = self.solver.solve()
        self.assertEqual(res, True)
        self.assertEqual(solution, (None, True))

    def test_3_long(self):
        self.solver.add_xor_clause([1, 2, 3], False)
        res, solution = self.solver.solve([1, 2])
        self.assertEqual(res, True)
        # self.assertEqual(solution, (None, True, True, False))

    def test_3_long2(self):
        self.solver.add_xor_clause([1, 2, 3], True)
        res, solution = self.solver.solve([1, -2])
        self.assertEqual(res, True)
        self.assertEqual(solution, (None, True, False, False))

    def test_long(self):
        for l in range(10, 30):
            self.setUp()
            toadd = []
            toassume = []
            solution_expected = [None]
            for i in range(1, l):
                toadd.append(i)
                solution_expected.append(False)
                if i != l - 1:
                    toassume.append(i * -1)

            self.solver.add_xor_clause(toadd, False)
            res, solution = self.solver.solve(toassume)
            self.assertEqual(res, True)
            self.assertEqual(solution, tuple(solution_expected))


class InitTester(unittest.TestCase):

    def test_wrong_args_to_solver(self):
        self.assertRaises(ValueError, Solver, threads=-1)
        self.assertRaises(ValueError, Solver, threads=0)
        self.assertRaises(ValueError, Solver, verbose=-1)
        self.assertRaises(ValueError, Solver, time_limit=-1)
        self.assertRaises(ValueError, Solver, confl_limit=-1)
        self.assertRaises(TypeError, Solver, threads="fail")
        self.assertRaises(TypeError, Solver, verbose="fail")
        self.assertRaises(TypeError, Solver, time_limit="fail")
        self.assertRaises(TypeError, Solver, confl_limit="fail")


class TestDump(unittest.TestCase):

    def setUp(self):
        self.solver = Solver()

    def test_max_glue_missing(self):
        self.assertRaises(TypeError,
                          self.solver.start_getting_small_clauses, 4)

    def test_one_dump(self):
        with open(_MODULE_DIR+"test.cnf", "r") as x:
            for line in x:
                line = line.strip()
                if "p" in line or "c" in line:
                    continue

                out = [int(x) for x in line.split()[:-1]]
                self.solver.add_clause(out)

        res, _ = self.solver.solve()
        self.assertEqual(res, True)

        self.solver.start_getting_small_clauses(4, max_glue=10)
        x = self.solver.get_next_small_clause()
        self.assertNotEquals(x, None)
        self.solver.end_getting_small_clauses()


class TestSolve(unittest.TestCase):

    def setUp(self):
        self.solver = Solver(threads=2)

    def test_wrong_args(self):
        self.assertRaises(TypeError, self.solver.add_clause, 'A')
        self.assertRaises(TypeError, self.solver.add_clause, 1)
        self.assertRaises(TypeError, self.solver.add_clause, 1.0)
        self.assertRaises(TypeError, self.solver.add_clause, object())
        self.assertRaises(TypeError, self.solver.add_clause, ['a'])
        self.assertRaises(
            TypeError, self.solver.add_clause, [[1, 2], [3, None]])
        self.assertRaises(ValueError, self.solver.add_clause, [1, 0])

    def test_no_clauses(self):
        for _ in range(7):
            self.assertEqual(self.solver.solve([]), (True, (None,)))

    def test_cnf1(self):
        for cl in clauses1:
            self.solver.add_clause(cl)
        res, solution = self.solver.solve()
        self.assertEqual(res, True)
        self.assertTrue(check_solution(clauses1, solution))

    def test_add_clauses(self):
        self.solver.add_clauses([[1], [-1]])
        res, solution = self.solver.solve()
        self.assertEqual(res, False)

    def test_add_clauses_wrong_zero(self):
        self.assertRaises(TypeError, self.solver.add_clause, [[1, 0], [-1]])

    def test_add_clauses_array_SAT(self):
        cls = array('i', [1, 2, 0, 1, 2, 0])
        self.solver.add_clauses(cls)
        res, solution = self.solver.solve()
        self.assertEqual(res, True)

    def test_add_clauses_array_UNSAT(self):
        cls = array('i', [-1, 0, 1, 0])
        self.solver.add_clauses(cls)
        res, solution = self.solver.solve()
        self.assertEqual(res, False)

    def test_add_clauses_array_unterminated(self):
        cls = array('i', [1, 2, 0, 1, 2])
        self.assertRaises(ValueError, self.solver.add_clause, cls)

    def test_bad_iter(self):
        class Liar:

            def __iter__(self):
                return None
        self.assertRaises(TypeError, self.solver.add_clause, Liar())

    def test_get_conflict(self):
        self.solver.add_clauses([[-1], [2], [3], [-4]])
        assume = [-2, 3, 4]

        res, model = self.solver.solve(assumptions=assume)
        self.assertEqual(res, False)

        confl = self.solver.get_conflict()
        self.assertEqual(isinstance(confl, list), True)
        self.assertNotIn(3, confl)

        if 2 in confl:
            self.assertIn(2, confl)
        elif -4 in confl:
            self.assertIn(-4, confl)
        else:
            self.assertEqual(False, True, msg="Either -2 or 4 should be conflicting!")

        assume = [2, 4]
        res, model = self.solver.solve(assumptions=assume)
        self.assertEqual(res, False)

        confl = self.solver.get_conflict()
        self.assertEqual(isinstance(confl, list), True)
        self.assertNotIn(2, confl)
        self.assertIn(-4, confl)

    def test_cnf2(self):
        for cl in clauses2:
            self.solver.add_clause(cl)
        self.assertEqual(self.solver.solve(), (False, None))

    def test_cnf3(self):
        for cl in clauses3:
            self.solver.add_clause(cl)
        res, solution = self.solver.solve()
        self.assertEqual(res, True)
        self.assertTrue(check_solution(clauses3, solution))

    def test_cnf1_confl_limit(self):
        for _ in range(1, 20):
            self.setUp()
            for cl in clauses1:
                self.solver.add_clause(cl)

            res, solution = self.solver.solve()
            self.assertTrue(res is None or check_solution(clauses1, solution))

    def test_by_re_curse(self):
        self.solver.add_clause([-1, -2, 3])
        res, _ = self.solver.solve()
        self.assertEqual(res, True)

        self.solver.add_clause([-5, 1])
        self.solver.add_clause([4, -3])
        self.solver.add_clause([2, 3, 5])
        res, _ = self.solver.solve()
        self.assertEqual(res, True)


class TestSolveTimeLimit(unittest.TestCase):

    def get_clauses(self):
        cls = []
        with open(_MODULE_DIR+"f400-r425-x000.cnf", "r") as f:
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                if line[0] == "p":
                    continue
                if line[0] == "c":
                    continue
                line = line.split()
                line = [int(l.strip()) for l in line]
                assert line[-1] == 0
                cls.append(line[:-1])

        return cls


    def test_time(self):
        SAT_TIME_LIMIT = 1
        clauses = self.get_clauses() #returns a few hundred short clauses
        t0 = time.time()
        solver = Solver(threads=4, time_limit=SAT_TIME_LIMIT)
        solver.add_clauses(clauses)
        sat, sol = solver.solve()
        took_time = time.time() - t0

        # NOTE: the above CNF solves in about 1 hour.
        # So anything below 10min is good. Setting 2s would work... no most
        # systems, but not on overloaded CI servers
        self.assertLess(took_time, 4)

# ------------------------------------------------------------------------


def run():
    print("sys.prefix: %s" % sys.prefix)
    print("sys.version: %s" % sys.version)
    try:
        print("pycryptosat version: %r" % pycryptosat.__version__)
    except AttributeError:
        pass
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestXor))
    suite.addTest(unittest.makeSuite(InitTester))
    suite.addTest(unittest.makeSuite(TestSolve))
    suite.addTest(unittest.makeSuite(TestDump))
    suite.addTest(unittest.makeSuite(TestSolveTimeLimit))

    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite)

    n_errors = len(result.errors)
    n_failures = len(result.failures)

    if n_errors or n_failures:
        print('\n\nSummary: %d errors and %d failures reported\n'%\
            (n_errors, n_failures))

    print()

    sys.exit(n_errors+n_failures)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import with_statement  # Required in 2.5
from __future__ import print_function
import debuglib
import sys

if len(sys.argv) != 4:
    print("You must give 3 arguments: input file, output file, seed")


debuglib.intersperse(sys.argv[1], sys.argv[2], sys.argv[3])
print("OK, output in %s" % sys.argv[2])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import sys
import re
import ntpath
import os

from optparse import OptionParser
usage = """
./reconf.py -n 15 ~/media/sat/out/satcomp091113/reconf0-09113-23-July-2015-mark-XZGKC-78499d3f2-tout-5000-mout-1600/*.stdout* ~/media/sat/out/satcomp091113/reconf14-091113-28-August-2015-mark-O5T8F-169dfc802-tout-5000-mout-1600/*.stdout*

NOTE: you *must* have reconf14 in there or a lot of data will be missing
in particular avg_confl_size, which will make the system error out on you.
"""
parser = OptionParser(usage=usage)
parser.add_option("-f", "--file",
                  dest="outfname", type=str, default="outfile",
                  help="print final values to this file")
parser.add_option("-n", "--num",
                  dest="num", type=int,
                  help="Number of reconfs")
parser.add_option("--dropdown",
                  dest="dropdown", type=float, default=0.02,
                  help="From traget 1.0 this is subtracted no matter what")
parser.add_option("--cutoff",
                  dest="cutoff", type=float, default=0.50,
                  help="At least this much or higher is needed for +")
parser.add_option("--divisor",
                  dest="divisor", type=float, default=1500.0,
                  help="Time difference is divided by this much and subtracted")
parser.add_option("--ignorethresh",
                  dest="ignore_threshold", type=float, default=4000.0,
                  help="If all solved above this score, ignore")
parser.add_option("--maxscore",
                  dest="maxscore", type=int, default=5000.0,
                  help="Scores go down from here")
parser.add_option("--ignore", "-i",
                  dest="ignore", type=str,
                  help="Ignore these reconfs")
parser.add_option("--verbose", "-r", dest="verbose", default=False,
                  action="store_true", help="More verbose")

(options, args) = parser.parse_args()
# print("files to parse are:", args)

ignore = {}
if options.ignore:
    for r in options.ignore.split(","):
        r = r.strip()
        r = int(r)
        ignore[r] = True

satzilla_feat_order = ["numClauses", "binary", "horn", "horn_mean", "horn_std", "horn_min", "horn_max", "horn_spread", "vcg_var_mean", "vcg_var_std", "vcg_var_min", "vcg_var_max", "vcg_var_spread", "vcg_cls_mean", "vcg_cls_std", "vcg_cls_min", "vcg_cls_max", "vcg_cls_spread", "pnr_var_mean", "pnr_var_std", "pnr_var_min", "pnr_var_max", "pnr_var_spread", "pnr_cls_mean", "pnr_cls_std", "pnr_cls_min", "pnr_cls_max", "pnr_cls_spread", "avg_confl_size", "confl_size_min", "confl_size_max", "avg_confl_glue", "confl_glue_min", "confl_glue_max", "avg_num_resolutions", "num_resolutions_min", "num_resolutions_max", "learnt_bins_per_confl", "avg_branch_depth", "branch_depth_min", "branch_depth_max", "avg_trail_depth_delta", "trail_depth_delta_min", "trail_depth_delta_max", "avg_branch_depth_delta", "props_per_confl", "confl_per_restart", "decisions_per_conflict", "irred_cl_distrib.glue_distr_mean", "irred_cl_distrib.glue_distr_var", "irred_cl_distrib.size_distr_mean", "irred_cl_distrib.size_distr_var", "irred_cl_distrib.uip_use_distr_mean", "irred_cl_distrib.uip_use_distr_var", "irred_cl_distrib.activity_distr_mean", "irred_cl_distrib.activity_distr_var", "red_cl_distrib.glue_distr_mean", "red_cl_distrib.glue_distr_var", "red_cl_distrib.size_distr_mean", "red_cl_distrib.size_distr_var", "red_cl_distrib.uip_use_distr_mean", "red_cl_distrib.uip_use_distr_var", "red_cl_distrib.activity_distr_mean", "red_cl_distrib.activity_distr_var"]

f = open("outs/reconf.names", "w")
f.write("reconf.                     | the target attribute\n\n")
f.write("name:                     label.\n")
for o in satzilla_feat_order:
    f.write("%s:                     continuous.\n" % o)
f.write("\nreconf:                 +,-.\n")
f.close()

if options.num is None:
    print("ERROR: You must give the number of reconfs")
    exit(-1)


def parse_satzilla_features_line(line):
    line = re.sub("c.*satzilla_features. ", "", line)
    line = line.strip().split(" ")
    dat = {}

    name = ""
    for elem, i in zip(line, range(1000)):
        elem = elem.strip(":").strip(",")
        if i % 2 == 0:
            name = elem
            continue

        dat[name] = elem
        name = ""
    return dat


def nobody_could_solve_it(reconf_score):
    for r_s_elem in reconf_score:
        if r_s_elem[1] != 0:
            return False

    return True


def all_above_fixed_score(reconf_score):
    for x in reconf_score:
        if x[0] in ignore:
            continue

        if x[1] < options.ignore_threshold:
            print("-> not ignoring, reconf %d is below ignore threshold" % x[0])
            return False

    return True


def print_satzilla_features_and_scores(fname, satzilla_features, reconfs_scores):
    r_s = sorted(reconfs_scores, key=lambda x: x[1])[::-1]
    best_reconf = r_s[0][0]
    best_reconf_score = r_s[0][1]
    print("r_s: ", r_s)

    if nobody_could_solve_it(r_s):
        print("Nobody could solve: %s" % fname)
        return -1, False

    if all_above_fixed_score(r_s):
        print("All above score: %s" % (fname))
        return -2, False

    print("Calculating +/- for %s" % fname)

    # calculate final array
    final_array = [0.0]*options.num
    val = 1.0
    best_score = r_s[0][1]
    for conf_score, i in zip(r_s, range(100)):
        diff = abs(conf_score[1]-best_score)
        best_score = conf_score[1]
        val -= float(diff)/options.divisor
        if diff > 0:
            val -= options.dropdown

        if val < 0.0 or conf_score[1] == 0:
            val = 0.0

        if conf_score[1] > 0:
            final_array[conf_score[0]] = val

    # assemble final string
    string = ""
    string += "%s," % fname
    for name in satzilla_feat_order:
        string += "%s," % satzilla_features[name]

    # print to console
    if True:
        string2 = str(string)
        string2 += "||"
        for a in final_array:
            string2 += "%.1f " % a

        print(string2)

    # print to files
    origstring = str(string)
    for i in range(options.num):
        # skip files we don't need to write to
        if i in ignore:
            continue

        string = str(origstring)
        if final_array[i] >= options.cutoff:
            string += "+"
        else:
            string += "-"

        outf[i].write(string + "\n")

    only_this_could_solve_it = r_s[1][1] == 0
    return best_reconf, only_this_could_solve_it


def parse_file(fname):
    f = open(fname, 'r')
    # print("fname orig:", fname)
    fname_clean = re.sub("cnf.gz-.*", "cnf.gz", fname)
    fname_clean = ntpath.split(fname_clean)[1]
    reconf = 0

    satisfiable = None
    satzilla_features = None
    score = 0
    for line in f:
        line = line.strip()
        #print("parsing line:", line)
        if satzilla_features is None and "satzilla_features" in line and "numClauses" in line:
            satzilla_features = parse_satzilla_features_line(line)

        if "Total time" in line:
            time_used = line.strip().split(":")[1].strip()
            score = int(round(float(time_used)))
            score = options.maxscore-score

        if "reconfigured" in line:
            reconf = line.split("to config")[1].strip()
            reconf = int(reconf)

        if "s SATIS" in line:
            satisfiable = True

        if "s UNSATIS" in line:
            satisfiable = False

    #if satisfiable == True:
    #    score = 0

    if reconf in ignore:
        score = 0

    # print("satzilla_features:", satzilla_features)
    return fname_clean, reconf, satzilla_features, score


all_files = set()
all_files_scores = {}
all_files_satzilla_features = {}
max_num_satzilla_features = 0
for x in args:
    print("# parsing infile:", x)
    fname, reconf, satzilla_features, score = parse_file(x)
    if fname in all_files:
        if all_files_satzilla_features[fname] != satzilla_features:
            print("different satzilla_features extracted for fname", fname)
            print("orig:", all_files_satzilla_features[fname])
            print("new: ", satzilla_features)
            print("Keeping the longer one!")

        if all_files_satzilla_features[fname] is None:
            num_satzilla_features = 0
        else:
            num_satzilla_features = len(all_files_satzilla_features[fname])

        if satzilla_features is not None and num_satzilla_features < len(satzilla_features):
            all_files_satzilla_features[fname] = satzilla_features
    else:
        all_files.add(fname)
        all_files_satzilla_features[fname] = satzilla_features
        all_files_scores[fname] = []

    #print("fname:", fname)
    all_files_scores[fname].append([reconf, score])

    sys.stdout.write(".")
    sys.stdout.flush()

print("END--------")
print("all files:", all_files)
print("")

outf = []
outfnames = []
for i in range(options.num):
    fname = options.outfname + str(i) + ".data"
    outfnames.append(fname)
    try:
        os.unlink(fname)
    except:
        pass

    if i not in ignore:
        outf.append(open(fname, "w"))
    else:
        outf.append(None)

best_reconf = {'all_above_fixed_score': 0, 'nobody_could_solve_it': 0}
for x in range(options.num):
    best_reconf[x] = 0
only_this = dict(best_reconf)


for fname in all_files:
    print("")
    print("calculating final DATs for CNF ", fname)
    if all_files_satzilla_features[fname] is None:
        print("-> solved too early, no satzilla_features, skipping")
        continue

    if options.verbose:
        print("-> all_files_satzilla_features[fname]:", all_files_satzilla_features[fname])
    if "avg_confl_size" not in all_files_satzilla_features[fname]:
        print("-> WARNING This is weird, probably not solved by one (different satzilla_features than everything else), skipping")
        continue

    if all_files_satzilla_features[fname] is None:
        print("-> satzilla_features for file is None: %s" % fname)

    if all_files_satzilla_features[fname] is not None:
        best, only_this_could_solve_it = print_satzilla_features_and_scores(fname, all_files_satzilla_features[fname], all_files_scores[fname])

        if best == -2:
            best = "all_above_fixed_score"

        if best == -1:
            best = "nobody_could_solve_it"

        print("-> best here:", best)
        best_reconf[best] = best_reconf[best] + 1
        if only_this_could_solve_it:
            only_this[best] = only_this[best] + 1

print("")
print("Wrote data files: %s\n" % outfnames)
print("\n-----------------------------")
print("best reconfs: ")
for a, b in best_reconf.items():
    if a not in ignore:
        print("%-20s : %-3d" % (a, b))

print("\n-----------------------------")
print("uniquely solved by: ")
for a, b in only_this.items():
    if a not in ignore:
        print("%-20s : %-3d" % (a, b))

for i in range(options.num):
    if outf[i] is not None:
        outf[i].close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import optparse
import time
import helper


class QueryFill (helper.QueryHelper):
    def __init__(self, dbfname):
        super(QueryFill, self).__init__(dbfname)

    def create_indexes(self):
        helper.drop_idxs(self.c)

        print("Creating needed indexes...")
        t = time.time()
        q = """
        create index `idxclid-del` on `cl_last_in_solver` (`clauseID`, `conflicts`);
        create index `idxclid-del2` on `used_clauses` (`clauseID`);
        create index `idxclid-del3` on `used_clauses` (`clauseID`, `used_at`);
        create index `idxclid1-2` on `clause_stats` (`clauseID`);
        """

        for l in q.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Creating index: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Index creation T: %-3.2f s" % (time.time() - t2))

        print("indexes created T: %-3.2f s" % (time.time() - t))

    def fill_sum_cl_use(self):
        print("Filling sum_cl_use...")

        t = time.time()
        q = """DROP TABLE IF EXISTS `sum_cl_use`;"""
        self.c.execute(q)
        q = """
        create table `sum_cl_use` (
            `clauseID` bigint(20) NOT NULL,
            `num_used` float(20) NOT NULL,
            `first_confl_used` bigint(20),
            `last_confl_used` bigint(20)
        );"""
        self.c.execute(q)
        print("sum_cl_use recreated T: %-3.2f s" % (time.time() - t))

        t = time.time()
        q = """insert into sum_cl_use
        (
        `clauseID`
        , `num_used`
        , `first_confl_used`
        , `last_confl_used`
        )
        select
        clauseID
        , sum(weight)
        , min(used_at)
        , max(used_at)
        from used_clauses as c
        group by clauseID;"""
        self.c.execute(q)
        print("sum_cl_use filled T: %-3.2f s" % (time.time() - t))

        t = time.time()
        q = """
        create index `idxclid21` on `sum_cl_use` (`clauseID`);
        """
        for l in q.split('\n'):
            self.c.execute(l)
        print("sum_cl_use indexes added T: %-3.2f s" % (time.time() - t))

        t = time.time()
        q = """
        insert into sum_cl_use
        (
        `clauseID`
        , `num_used`
        , `first_confl_used`
        , `last_confl_used`
        )
        select
        clstats.clauseID -- `clauseID`
        , 0     --  `num_used`,
        , NULL   --  `first_confl_used`
        , NULL   --  `last_confl_used`
        from clause_stats as clstats left join sum_cl_use
        on clstats.clauseID = sum_cl_use.clauseID
        where
        sum_cl_use.clauseID is NULL
        and clstats.clauseID != 0;
        """
        self.c.execute(q)
        print("sum_cl_use added bad claues T: %-3.2f s" % (time.time() - t))


if __name__ == "__main__":
    usage = "usage: %prog [options] sqlitedb"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")

    (options, args) = parser.parse_args()

    if len(args) < 1:
        print("ERROR: You must give the sqlite file!")
        exit(-1)

    with QueryFill(args[0]) as q:
        q.create_indexes()
        q.fill_sum_cl_use()
        helper.drop_idxs(q.c)

    print("Done.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


from __future__ import print_function
import sys

from optparse import OptionParser
parser = OptionParser()
parser.add_option("-n", "--num",
                  dest="num", type=int,
                  help="Number of reconfs")
parser.add_option("--ignore", "-i",
                  dest="ignore", type=str,
                  help="Ignore these reconfs")

(options, args) = parser.parse_args()

ignore = {}
if options.ignore:
    for r in options.ignore.split(","):
        r = r.strip()
        r = int(r)
        ignore[r] = True

if options.num is None:
    print("ERROR: You must give the number of reconfs")
    exit(-1)

print("""/******************************************
Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
***********************************************/

#include "satzilla_features.h"
#include <iostream>
using std::cout;
using std::endl;

namespace CMSat {
""")

for i in range(options.num):
    if i not in ignore:
        print("double get_score%d(const SatZillaFeatures& satzilla_feat, const int verb);" % i)

print("""
int get_reconf_from_satzilla_features(const SatZillaFeatures& satzilla_feat, const int verb)
{
\tdouble best_score = 0.0;
\tint best_val = 0;
\tdouble score;
""")

for i in range(options.num):
    if i in ignore:
        continue

    print("""
\tscore = get_score%d(satzilla_feat, verb);
\tif (verb >= 2)
\t\tcout << "c Score for reconf %d is " << score << endl;
\tif (best_score < score) {
\t\tbest_score = score;
\t\tbest_val = %d;
\t}
""" % (i, i, i))

print("""
\tif (verb >= 2)
\t\tcout << "c Winning reconf is " << best_val << endl;
\treturn best_val;
}

""")


def read_one_reconf(reconf_num):
    sys.stderr.write("Parsing reconf num %d\n" % reconf_num)
    f = open("outs/out%d.rules" % reconf_num)
    num_conds = 0
    cond_no = 0
    num_rules = 0
    rule_no = 0
    string = ""

    print("""
double get_score%d(const SatZillaFeatures& satzilla_feat, const int verb)
{""" % reconf_num)
    for line in f:
        if "id=" in line:
            continue

        line = line.strip()
        line = line.split(" ")
        dat = {}
        for elem in line:
            elems = elem.split("=")
            elems = [e.strip("\"") for e in elems]
            dat[elems[0]] = elems[1]

        if "conds" in dat:
            assert num_conds == cond_no
            num_conds = int(dat["conds"])
            rule_class = dat["class"]
            cond_no = 0
            confidence = float(dat["confidence"])
            continue

        if "entries" in dat:
            continue

        if "rules" in dat:
            num_rules = int(dat["rules"])

        if "default" in dat:
            default = dat["default"]
            if default == "+":
                print("\tdouble default_val = %.2f;\n" % (1.0))
            else:
                print("\tdouble default_val = %.2f;\n" % (0.0))

            print("\tdouble total_plus = 0.0;")
            print("\tdouble total_neg = 0.0;")
            continue

        # process rules
        if cond_no == 0:
            string = "\tif ("
        else:
            string += " &&\n\t\t"

        string += "(satzilla_feat.%s %s %.5f)" % (dat["att"], dat["result"],
                                         float(dat["cut"]))
        cond_no += 1

        # end rules
        if cond_no == num_conds:
            string += ")\n\t{"
            print(string)

            string = ""
            if rule_class == "+":
                string += "\t\ttotal_plus += %.3f;" % confidence
            else:
                string += "\t\ttotal_neg += %.3f;" % confidence

            print(string)
            print("\t}")
            rule_no += 1

    print("\t// num_rules:", num_rules)
    print("\t// rule_no:", rule_no)
    sys.stderr.write("num rules: %s rule_no: %s\n" % (num_rules, rule_no))
    assert num_rules == rule_no
    print("\t// default is:", default)
    print("""
\tif (total_plus == 0.0 && total_neg == 0.0) {
\t\treturn default_val;
\t}
\tif (verb >= 2) {
\t\t//cout << "c plus: " << total_plus << " , neg: " << total_neg << endl;
\t}
\treturn total_plus - total_neg;
}
""")


for i in range(options.num):
    if i not in ignore:
        read_one_reconf(i)

print("""
} //end namespace""")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import optparse
import random
import sys

lpnfile = sys.argv[1]
satout = sys.argv[2]

sat = None
sol = {}
with open(satout, "r") as f:
    for line in f:
        line = line.strip()
        if len(line) < 1:
            continue

        if line[0] == "s":
            if "s SATISFIABLE" in line:
                sat = True
                continue
            if "s UNSATISFIABLE" in line:
                sat = False
                continue
            print("ERROR: Nether SAT, nor UNSAT on solution line!")
            exit(-1)

        if line[0] == "v":
            line=line.strip()
            line=line[1:]
            for lit in line.split():
                lit = int(lit)
                var = abs(lit)
                if var == 0:
                    continue
                sol[var] = lit > 0

if sat is None:
    print("ERRROR could not recover solution from SAT output")
    exit(-1)

# print("Recovered SAT: ", sat)
# print("Recovered Solution : ", sol)

corr_sat = None
corr_sol = {}
with open(lpnfile, "r") as f:
    for line in f:
        line=line.strip()
        if "correct output" not in line:
            continue
        if "UNSAT" in line:
            corr_sat = False
            break
        corr_sat = True
        assert "SAT" in line

        line = line.split(":")[1]
        for lit in line.split()[1:]: # first is text "SAT"
            lit = int(lit)
            var = abs(lit)
            corr_sol[var] = lit > 0

if corr_sat is None:
    print("ERROR: could not recover correct solution from problem file")
    exit(-1)

if corr_sat is False:
    if sat == False:
        print("UNSAT both, all good!")
        exit(0)
    if sat == True:
        print("ERROR: buuuug!!! Should be UNSAT but it's SAT!!")
        exit(-1)

assert corr_sat == True
if sat == False:
    print("ERROR: buuuug!!! Should be SAT but it's UNSAT!!")
    exit(-1)

for v,val in corr_sol.items():
    if v not in sol:
        print("ERROR: Var %d is in problem, but not in solution" % v)
        exit(-1)
    if sol[v] != corr_sol[v]:
        print("ERROR: Var %d is %s in problem, but %s in solution" % (v, corr_sol[v], sol[v]))
        exit(-1)
    assert sol[v] == corr_sol[v]

print("OK, SAT, checked all solution values") #: %s" % corr_sol)
exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
import sqlite3
import optparse
import operator
import time
import helper


class Query:
    def __init__(self, dbfname):
        self.conn = sqlite3.connect(dbfname)
        self.c = self.conn.cursor()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def create_indexes(self):
        helper.drop_idxs(self.c)

        print("Recreating indexes...")
        queries = """
        create index `idx1` on `tags` (`runid`, `name`);
        create index `idx2` on `timepassed` (`runid`, `elapsed`);
        create index `idx3` on `timepassed` (`runid`, `elapsed`, `name`);
        create index `idx4` on `memused` (`runid`, `MB`, `name`);

        """
        for l in queries.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Creating index: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Index creation T: %-3.2f s" % (time.time() - t2))

        print("indexes created T: %-3.2f s" % (time.time() - t))

    def find_time_outliers(self):
        print("----------- TIME OUTLIERS --------------")
        query = """
        select tags.val, timepassed.name, timepassed.elapsed
        from timepassed,tags
        where
        timepassed.name != 'search'
        and timepassed.elapsed > %d
        and tags.name="filename"
        and tags.runid = timepassed.runid

        order by elapsed desc;
        """ % (options.maxtime)

        for row in self.c.execute(query):
            operation = row[1]
            t = row[2]
            fname = self.get_fname(row[0])
            print("%-32s    %-20s   %.1fs" % (fname, operation, t))

    def check_memory(self):
        print("----------- MEMORY OUTLIERS --------------")

        query = """
        select tags.val, memused.name, max(memused.MB)
        from memused,tags
        where
        tags.name="filename"
        and memused.MB > %d
        and memused.name != 'vm'
        and memused.name != 'rss'
        and memused.name != 'longclauses'
        and tags.runid = memused.runid

        group by tags.val, memused.name
        order by MB desc;
        """ % (options.maxmemory)

        for row in self.c.execute(query):
            subsystem = row[1]
            gigs = row[2]/1000.0
            fname = self.get_fname(row[0])
            print("%-32s    %-20s   %.1f GB" % (fname, subsystem, gigs))

    def check_memory_rss(self):
        print("----------- MEMORY OUTLIERS RSS --------------")

        query = """
        select tags.val, memused.name, max(memused.MB)
        from memused,tags
        where
        tags.name="filename"
        and memused.MB > %d
        and memused.name == 'rss'
        and tags.runid = memused.runid

        group by tags.val, memused.name
        order by MB desc;
        """ % (options.maxmemory*2)

        for row in self.c.execute(query):
            subsystem = row[1]
            gigs = row[2]/1000.0
            fname = self.get_fname(row[0])
            print("%-32s    %-20s   %.1f GB" % (fname, subsystem, gigs))

    def find_worst_unaccounted_memory(self):
        print("----------- Largest RSS vs counted differences --------------")
        query = """
        select tags.val, a.`runtime`, abs((b.rss-a.mysum)/b.rss) as differperc,
            a.mysum as counted, b.rss as total
        from tags,

        (select runid, `runtime`, sum(MB) as mysum
        from memused
        where name != 'rss'
        and name != 'vm'
        group by `runtime`) as a,

        (select runid, name, `runtime`, MB as rss
        from memused
        where name = 'rss') as b

        where
        tags.name="filename"
        and a.`runtime` = b.`runtime`
        and total > %d
        and a.runid = tags.runid
        and b.runid = tags.runid

        order by differperc desc
        """ % (options.minmemory)

        for row,_ in zip(self.c.execute(query), range(10)):
            t = row[1]
            diff_perc = row[2]
            counted = row[3]
            total = row[4]
            fname = self.get_fname(row[0])
            print("%-32s    at: %-8.1fs   counted: %3.2f GB   rss: %3.2f GB" % (fname, t, counted/1000.0, total/1000.0))

    def memory_distrib(self):
        print("----------- MEMORY DISTRIBUTION --------------")
        print("all divided by RSS -- resident size")

        query = """
        select sum(MB)
        from memused
        where name != 'vm'
        and name != 'rss';
        """
        for row in self.c.execute(query):
            recorded_mem = float(row[0])

        query = """
        select sum(MB)
        from memused
        where name = 'rss';
        """
        for row in self.c.execute(query):
            rss_mem = float(row[0])

        unaccounted = rss_mem - recorded_mem;
        print("%-20s   %.1f%%" % ("unaccounted", unaccounted/rss_mem*100.0))

        query = """
        select name, sum(MB) as memsum
        from memused
        where name != 'vm'
        and name != 'rss'
        group by name
        order by memsum desc;
        """

        for row in self.c.execute(query):
            subsystem = row[0]
            mbs = float(row[1])
            print("%-20s   %.1f%%" % (subsystem, mbs/rss_mem*100.0))

    def get_fname(self, val) :
        fname = val.split("/")
        fname = fname[len(fname)-1]
        fname = fname.rstrip(".cnf.gz")
        #if len(fname) > 40:
        #    fname = fname[:30] + "..." + fname[len(fname)-10:]

        return fname

    def calc_time_spent(self):
        print("----------- TIME DISTRIBUTION --------------")

        query = """
        select sum(elapsed)
        from timepassed
        where name='search';
        """
        for row in self.c.execute(query):
            search_time = float(row[0])

        query = """
        select sum(elapsed)
        from timepassed
        where name!='search';
        """
        for row in self.c.execute(query):
            other_time = float(row[0])

        total = search_time + other_time
        print("Total: %10.1fh Search: %3.1f%%, Other: %3.1f%%" %
              (total/3600, search_time/total*100, other_time/total*100))

        query = """
        select name, sum(elapsed)
        from timepassed
        group by name;
        """
        times = {}
        for row in self.c.execute(query):
            times[row[0]] = float(row[1])

        # print("Names: %s" % times)
        sorted_t = sorted(times.items(), key=operator.itemgetter(1), reverse=True)
        for (name, t) in sorted_t:
            name = name[:40]
            print("%-40s   %3.1f%%" % (name, t/total*100))

    def find_intersting_problems(self):
        print("----------- Interesting problems for learning --------------")

        #Find CNFs that are interesting:
        #* solved in under 500s
        #* UNSAT
        #* more conflicts than 60'000

        #last conflict > 60000, UNSAT, solvetime under 500s
        query = """
        select tags.val, a.maxtime, a.maxconfl, mems.maxmem

        from
        (select runid, max(conflicts) as maxconfl, max(`runtime`) as maxtime
        from timepassed
        ) as a

        , (select *
        from finishup
        where status = "l_False"
        ) as b

        , (select runid, max(MB) as maxmem
        from memused
        ) as mems

        , tags

        where a.maxconfl > 20000
        and a.maxconfl < 400000
        and a.maxtime < 400
        and a.maxtime > 10
        and tags.name = "filename"
        and tags.runid = a.runid
        and tags.runid = b.runid


        order by maxtime desc
        """

        for row in self.c.execute(query):
            fname = row[1].split("/")
            fname = fname[len(fname)-1]
            t = row[1]
            confl = row[2]
            mb = row[3]
            print("t(mins): %-6.1f  confl(K): %-6.1f  mem(GB): %-6.1f  fname: %s" %
                  (t/60.0, confl/(1000.0), mb/1024.0, fname))


if __name__ == "__main__":
    usage = "usage: %prog [options] sqlitedb"
    parser = optparse.OptionParser(usage=usage)

    parser.add_option("--createind", action="store_true", default=False,
                      dest="create_indexes", help="Create indexes")

    parser.add_option("--maxtime", metavar="CUTOFF",
                      dest="maxtime", default=20, type=int,
                      help="Max time for an operation")

    parser.add_option("--maxmemory", metavar="CUTOFF",
                      dest="maxmemory", default=500, type=int,
                      help="Max memory for a subsystem")

    parser.add_option("--minmemory", metavar="MINMEM",
                      dest="minmemory", default=800, type=int,
                      help="Minimum memory to be checked for RSS vs counted check")

    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")

    (options, args) = parser.parse_args()

    if len(args) != 1:
        print("ERROR: You must give exactly one argument, the sqlite3 database file")
        exit(-1)

    dbfname = args[0]
    print("Using sqlite3db file %s" % dbfname)

    #peform queries
    with Query(dbfname) as q:
        if options.create_indexes:
            q.create_indexes()
        q.find_intersting_problems()
        q.find_worst_unaccounted_memory()
        q.check_memory_rss()
        q.check_memory()
        q.memory_distrib()
        q.find_time_outliers()
        q.calc_time_spent()



#select timepassed.runID, tag,elapsed from timepassed,tags where name like 'shorten and str red%' and elapsed > 2 and tags.runID = timepassed.runID;

#select * from timepassed where elapsed > 20 and name not like 'search';

#select * from startup, solverRun, finishup where finishup.runID = solverRun.runID and (finishup.endTime - startup.startTime) < 30 and solverRun.version = "618b5e79fdd8a15918e19fb76ca08aa069f14b54" and solverRun.runID = startup.runID;
